#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for dynamic utility methods.
"""

import pytest
from pyomo.environ import (
    ConcreteModel,
    Block,
    Constraint,
    Var,
    Set,
    TransformationFactory,
)
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.common.collections import ComponentSet
import idaes.logger as idaeslog
from idaes.core.util.dyn_utils import *

__author__ = "Robert Parker"


@pytest.mark.unit
def test_fix_and_deactivate():
    m = ConcreteModel()
    m.time = ContinuousSet(bounds=(0, 10))
    m.space = ContinuousSet(bounds=(0, 5))
    m.set1 = Set(initialize=["a", "b", "c"])
    m.set2 = Set(initialize=["d", "e", "f"])
    m.fs = Block()

    m.fs.v0 = Var(m.space, initialize=1)

    @m.fs.Block()
    def b1(b):
        b.v = Var(m.time, m.space, initialize=1)
        b.dv = DerivativeVar(b.v, wrt=m.time)

        b.con = Constraint(
            m.time, m.space, rule=lambda b, t, x: b.dv[t, x] == 7 - b.v[t, x]
        )

        @b.Block(m.time)
        def b2(b, t):
            b.v = Var(initialize=2)

    @m.fs.Block(m.time, m.space)
    def b2(b, t, x):
        b.v = Var(m.set1, initialize=2)

        @b.Block(m.set1)
        def b3(b, c):
            b.v = Var(m.set2, initialize=3)

            @b.Constraint(m.set2)
            def con(b, s):
                return 5 * b.v[s] == m.fs.b2[m.time.first(), m.space.first()].v[c]

    @m.fs.Constraint(m.time)
    def con1(fs, t):
        return fs.b1.v[t, m.space.last()] == 5

    @m.fs.Constraint(m.space)
    def con2(fs, x):
        return fs.b1.v[m.time.first(), x] == fs.v0[x]

    disc = TransformationFactory("dae.collocation")
    disc.apply_to(m, wrt=m.time, nfe=5, ncp=2, scheme="LAGRANGE-RADAU")
    disc.apply_to(m, wrt=m.space, nfe=5, ncp=2, scheme="LAGRANGE-RADAU")

    for t in m.time:
        m.fs.b1.v[t, m.space.first()].fix()

    active_dict = get_activity_dict(m.fs)
    for comp in m.fs.component_data_objects(Constraint, Block):
        assert active_dict[id(comp)] is True

    deactivate_model_at(m, m.time, m.time.at(2))
    assert m.fs.con1[m.time.at(1)].active
    assert not m.fs.con1[m.time.at(2)].active
    assert m.fs.con2[m.space.at(1)].active
    assert not m.fs.b1.con[m.time.at(2), m.space.at(1)].active
    assert not m.fs.b2[m.time.at(2), m.space.last()].active
    assert m.fs.b2[m.time.at(2), m.space.last()].b3["a"].con["e"].active

    deactivate_model_at(m, m.time, [m.time.at(1), m.time.at(3)], outlvl=idaeslog.ERROR)
    # Higher outlvl threshold as will encounter warning trying to deactivate
    # disc equations at time.first()
    assert not m.fs.con1[m.time.at(1)].active
    assert not m.fs.con1[m.time.at(3)].active
    assert not m.fs.b1.con[m.time.at(1), m.space.at(1)].active
    assert not m.fs.b1.con[m.time.at(3), m.space.at(1)].active

    init_derivs = get_derivatives_at(m, m.time, m.time.first())[m.time.first()]
    init_deriv_names = [var.name for var in init_derivs]

    assert m.fs.b1.dv[m.time.first(), m.space.first()] in init_derivs
    assert m.fs.b1.dv[m.time.at(1), m.space.at(1)].name in init_deriv_names

    deriv_dict = get_derivatives_at(m, m.time, [m.time.first(), m.time.last()])
    deriv_name_dict = {t: [d.name for d in deriv_dict[t]] for t in deriv_dict.keys()}
    assert m.time.first() in deriv_name_dict.keys()
    assert m.time.last() in deriv_name_dict.keys()
    assert (
        m.fs.b1.dv[m.time.last(), m.space.at(1)].name in deriv_name_dict[m.time.last()]
    )
    assert (
        m.fs.b1.dv[m.time.last(), m.space.at(1)].name
        not in deriv_name_dict[m.time.first()]
    )

    vars_unindexed = fix_vars_unindexed_by(m, m.time)
    cons_unindexed = deactivate_constraints_unindexed_by(m, m.time)

    unindexed_vars = ComponentSet(vars_unindexed)
    assert m.fs.v0[m.space.at(1)] in unindexed_vars
    assert m.fs.b1.b2[m.time.at(1)].v not in unindexed_vars
    assert m.fs.con2[m.space.at(2)] in cons_unindexed
    assert m.fs.con1[m.time.at(1)] not in cons_unindexed
    assert not m.fs.con2[m.space.at(1)].active
    assert m.fs.v0[m.space.at(1)].fixed

    path = path_from_block(
        m.fs.b2[m.time.at(1), m.space.at(1)].b3["a"].v, m, include_comp=False
    )
    assert path == [("fs", None), ("b2", (m.time.at(1), m.space.at(1))), ("b3", "a")]
    path = path_from_block(
        m.fs.b2[m.time.at(1), m.space.at(1)].b3["a"].v, m, include_comp=True
    )
    assert path == [
        ("fs", None),
        ("b2", (m.time.at(1), m.space.at(1))),
        ("b3", "a"),
        ("v", None),
    ]
    path = path_from_block(
        m.fs.b2[m.time.at(1), m.space.at(1)].b3["a"].v["f"], m, include_comp=True
    )
    assert path == [
        ("fs", None),
        ("b2", (m.time.at(1), m.space.at(1))),
        ("b3", "a"),
        ("v", "f"),
    ]
    path = path_from_block(
        m.fs.b2[m.time.at(1), m.space.at(1)].b3["a"].v["f"],
        m.fs.b2[m.time.at(1), m.space.at(1)],
        include_comp=True,
    )
    assert path == [("b3", "a"), ("v", "f")]
    path = path_from_block(m.fs.b1.con[m.time.at(1), m.space.at(1)], m.fs)
    assert path == [("b1", None)]

    m.fs.b1.b2[m.time.at(1)].v.set_value(-1)
    for x in m.space:
        m.fs.b1.v[m.time.at(1), x].set_value(-1)
        m.fs.b1.dv[m.time.at(1), x].set_value(-1)
        for c1 in m.set1:
            m.fs.b2[m.time.at(1), x].v[c1].set_value(-1)
            for c2 in m.set2:
                m.fs.b2[m.time.at(1), x].b3[c1].v[c2].set_value(-1)

    copy_values_at_time(m, m, m.time.last(), m.time.at(1), copy_fixed=False)
    assert m.fs.b1.b2[m.time.last()].v.value == -1
    for x in m.space:
        if x != m.space.first():
            assert m.fs.b1.v[m.time.last(), x].value == -1
        else:
            assert m.fs.b1.v[m.time.last(), x].value == 1
            assert m.fs.b1.v[m.time.last(), x].fixed
        assert m.fs.b1.dv[m.time.last(), x].value == -1
        for c1 in m.set1:
            assert m.fs.b2[m.time.last(), x].v[c1].value == -1
            for c2 in m.set2:
                assert m.fs.b2[m.time.at(1), x].b3[c1].v[c2].value == -1


@pytest.mark.unit
def test_copy_non_time_indexed_values():
    m1 = ConcreteModel()
    m1.time = Set(initialize=[1, 2, 3, 4, 5])
    m1.v1 = Var(m1.time, initialize=1)
    m1.v2 = Var(initialize=1)

    @m1.Block(["a", "b"])
    def b1(b, i):
        b.v3 = Var(initialize=1)

        @b.Block(m1.time)
        def b2(b, t):
            b.v4 = Var(initialize=1)

        @b.Block()
        def b4(b):
            b.v6 = Var(initialize=1)

    @m1.Block(m1.time)
    def b3(b, t):
        b.v5 = Var(initialize=1)

    m2 = ConcreteModel()
    m2.time = Set(initialize=[1, 2, 3, 4, 5])
    m2.v1 = Var(m2.time, initialize=2)
    m2.v2 = Var(initialize=2)

    @m2.Block(["a", "b"])
    def b1(b):
        b.v3 = Var(initialize=2)

        @b.Block(m2.time)
        def b2(b, t):
            b.v4 = Var(initialize=2)

        @b.Block()
        def b4(b):
            b.v6 = Var(initialize=2)

    @m2.Block(m1.time)
    def b3(b, t):
        b.v5 = Var(initialize=2)

    copy_non_time_indexed_values(m1, m2)
    assert m1.v1[1].value != m2.v1[1].value
    assert m1.v2.value == m2.v2.value == 2
    assert m1.b1["a"].v3.value == m2.b1["a"].v3.value == 2
    assert m1.b1["b"].b4.v6.value == m2.b1["b"].b4.v6.value == 2
    assert m1.b3[3].v5.value != m2.b3[3].v5.value


@pytest.mark.unit
def test_find_comp_in_block():
    m1 = ConcreteModel()

    @m1.Block([1, 2, 3])
    def b1(b):
        b.v = Var([1, 2, 3])

    m2 = ConcreteModel()

    @m2.Block([1, 2, 3])
    def b1(b):
        b.v = Var([1, 2, 3, 4])

    @m2.Block([1, 2, 3])
    def b2(b):
        b.v = Var([1, 2, 3])

    v1 = m1.b1[1].v[1]

    assert find_comp_in_block(m2, m1, v1) is m2.b1[1].v[1]

    v2 = m2.b2[1].v[1]
    v3 = m2.b1[3].v[4]

    # These should result in Attribute/KeyErrors

    with pytest.raises(AttributeError, match=r".*has no attribute.*"):
        find_comp_in_block(m1, m2, v2)
    with pytest.raises(KeyError, match=r".*is not a valid index.*"):
        find_comp_in_block(m1, m2, v3)
    assert find_comp_in_block(m1, m2, v2, allow_miss=True) is None
    assert find_comp_in_block(m1, m2, v3, allow_miss=True) is None


@pytest.mark.unit
def test_find_comp_in_block_at_time():
    m1 = ConcreteModel()
    m1.time = Set(initialize=[1, 2, 3])

    @m1.Block(m1.time)
    def b1(b):
        b.v = Var(m1.time)

    @m1.Block([1, 2, 3])
    def b(bl):
        bl.v = Var(m1.time)
        bl.v2 = Var(m1.time, ["a", "b", "c"])

    m2 = ConcreteModel()
    m2.time = Set(initialize=[1, 2, 3, 4, 5, 6])

    @m2.Block(m2.time)
    def b1(b):
        b.v = Var(m2.time)

    @m2.Block([1, 2, 3, 4])
    def b(bl):
        bl.v = Var(m2.time)
        bl.v2 = Var(m2.time, ["a", "b", "c"])

    @m2.Block([1, 2, 3])
    def b2(b):
        b.v = Var(m2.time)

    v1 = m1.b1[1].v[1]
    v3 = m1.b[3].v[1]

    assert find_comp_in_block_at_time(m2, m1, v1, m2.time, 4) is m2.b1[4].v[4]
    assert find_comp_in_block_at_time(m2, m1, v3, m2.time, 5) is m2.b[3].v[5]

    v = m1.b[1].v2[1, "a"]
    assert find_comp_in_block_at_time(m2, m1, v, m2.time, 6) is m2.b[1].v2[6, "a"]

    v2 = m2.b2[1].v[1]
    v4 = m2.b[4].v[1]

    # Should result in exceptions:
    with pytest.raises(AttributeError, match=r".*has no attribute.*"):
        find_comp_in_block_at_time(m1, m2, v2, m1.time, 3)
    with pytest.raises(KeyError, match=r".*is not a valid index.*"):
        find_comp_in_block_at_time(m1, m2, v4, m1.time, 3)

    assert find_comp_in_block_at_time(m1, m2, v2, m1.time, 3, allow_miss=True) is None
    assert find_comp_in_block_at_time(m1, m2, v4, m1.time, 3, allow_miss=True) is None


@pytest.mark.unit
def test_get_location_of_coordinate_set():
    m = ConcreteModel()
    m.s1 = Set(initialize=[1, 2, 3])
    m.s2 = Set(initialize=[("a", 1), ("b", 2)])
    m.s3 = Set(initialize=[("a", 1, 0), ("b", 2, 1)])
    m.v1 = Var(m.s1)
    m.v2 = Var(m.s1, m.s2)
    m.v121 = Var(m.s1, m.s2, m.s1)
    m.v3 = Var(m.s3, m.s1, m.s2)

    assert get_location_of_coordinate_set(m.v1.index_set(), m.s1) == 0
    assert get_location_of_coordinate_set(m.v2.index_set(), m.s1) == 0
    assert get_location_of_coordinate_set(m.v3.index_set(), m.s1) == 3

    # These should each raise a ValueError
    with pytest.raises(ValueError):
        get_location_of_coordinate_set(m.v1.index_set(), m.s2)
    with pytest.raises(ValueError):
        get_location_of_coordinate_set(m.v121.index_set(), m.s1)


@pytest.mark.unit
def test_get_index_of_set():
    m = ConcreteModel()
    m.s1 = Set(initialize=[1, 2, 3])
    m.s2 = Set(initialize=[("a", 1), ("b", 2)])
    m.s3 = Set(initialize=[("a", 1, 0), ("b", 2, 1)])
    m.v0 = Var()
    m.v1 = Var(m.s1)
    m.v2 = Var(m.s1, m.s2)
    m.v121 = Var(m.s1, m.s2, m.s1)
    m.v3 = Var(m.s3, m.s1, m.s2)

    assert get_index_of_set(m.v1[2], m.s1) == 2
    assert get_index_of_set(m.v2[2, "a", 1], m.s1) == 2
    assert get_index_of_set(m.v3["b", 2, 1, 3, "b", 2], m.s1) == 3

    with pytest.raises(ValueError) as exc_test:
        get_index_of_set(m.v0, m.s1)
    with pytest.raises(ValueError) as exc_test:
        get_index_of_set(m.v2[1, "a", 1], m.s2)
    with pytest.raises(ValueError) as exc_test:
        get_index_of_set(m.v2[1, "b", 2], m.s3)


@pytest.mark.unit
def test_get_implicit_index_of_set():
    m = ConcreteModel()
    m.s1 = Set(initialize=[1, 2, 3])
    m.s2 = Set(initialize=["a", "b", "c"])
    m.s3 = Set(initialize=[("d", 4), ("e", 5)])

    @m.Block()
    def b1(b):
        @b.Block(m.s3, m.s1)
        def b2(b):
            @b.Block()
            def b3(b):
                b.v1 = Var(m.s2)
                b.v2 = Var(m.s1)

    assert get_implicit_index_of_set(m.b1.b2["d", 4, 1].b3.v1["a"], m.s1) == 1
    assert get_implicit_index_of_set(m.b1.b2["d", 4, 1].b3.v1["a"], m.s2) == "a"

    with pytest.raises(ValueError) as exc_test:
        get_implicit_index_of_set(m.b1.b2["e", 5, 2].b3.v2[1], m.s1)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Test for element-by-element initialization on PID controller. This is important
as PID controllers have additional time-linking variables beyond derivative
and differential variables.
"""

import pytest
from pyomo.environ import (
    Block,
    ConcreteModel,
    Constraint,
    Var,
    TransformationFactory,
    units as pyunits,
)
from pyomo.network import Arc
from pyomo.common.collections import ComponentMap

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.unit_models import CSTR, Mixer, MomentumMixingType
from idaes.models.control.controller import (
    PIDController,
    ControllerType,
    ControllerMVBoundType,
)
from idaes.core.util.initialization import initialize_by_time_element
from idaes.core.util.tests.test_initialization import (
    AqueousEnzymeParameterBlock,
    EnzymeReactionParameterBlock,
)
from idaes.core.solvers import get_solver

import idaes.logger as idaeslog

__author__ = "Robert Parker"


solver = get_solver("ipopt_v2")


def make_model(horizon=6, ntfe=60, ntcp=2, inlet_E=11.91, inlet_S=12.92):

    time_set = [0, horizon]

    m = ConcreteModel(name="CSTR with level control")
    m.fs = FlowsheetBlock(dynamic=True, time_set=time_set, time_units=pyunits.s)

    m.fs.properties = AqueousEnzymeParameterBlock()
    m.fs.reactions = EnzymeReactionParameterBlock(property_package=m.fs.properties)
    m.fs.cstr = CSTR(
        has_holdup=True,
        property_package=m.fs.properties,
        reaction_package=m.fs.reactions,
        material_balance_type=MaterialBalanceType.componentTotal,
        energy_balance_type=EnergyBalanceType.enthalpyTotal,
        momentum_balance_type=MomentumBalanceType.none,
        has_heat_of_reaction=True,
    )
    # MomentumBalanceType.none used because the property package doesn't
    # include pressure.

    m.fs.mixer = Mixer(
        property_package=m.fs.properties,
        material_balance_type=MaterialBalanceType.componentTotal,
        momentum_mixing_type=MomentumMixingType.none,
        num_inlets=2,
        inlet_list=["S_inlet", "E_inlet"],
    )
    # Allegedly the proper energy balance is being used...

    m.fs.pid = PIDController(
        process_var=m.fs.cstr.volume,
        manipulated_var=m.fs.cstr.outlet.flow_vol,
        mv_bound_type=ControllerMVBoundType.SMOOTH_BOUND,
        calculate_initial_integral=True,
        # ^ Why would initial integral be calculated
        # to be nonzero?
        controller_type=ControllerType.PID,
    )

    # Time discretization
    disc = TransformationFactory("dae.collocation")
    disc.apply_to(m, wrt=m.fs.time, nfe=ntfe, ncp=ntcp, scheme="LAGRANGE-RADAU")

    m.fs.pid.gain_p.fix(-1.0)
    m.fs.pid.gain_i.fix(10)
    m.fs.pid.gain_d.fix(0.0)
    m.fs.pid.setpoint.fix(1.0)
    m.fs.pid.mv_lb.set_value(0.5)
    m.fs.pid.mv_ub.set_value(5.0)
    m.fs.pid.mv_ref.fix(0.0)
    m.fs.pid.derivative_term[m.fs.time.first()].fix(0)

    # Fix initial condition for volume:
    m.fs.cstr.volume.unfix()
    m.fs.cstr.volume[m.fs.time.first()].fix(1.0)

    # Fix initial conditions for other variables:
    for p, j in m.fs.properties.phase_list * m.fs.properties.component_list:
        if j == "Solvent":
            continue
        m.fs.cstr.control_volume.material_holdup[0, p, j].fix(0.001)
    # Note: Model does not solve when initial conditions are empty tank
    m.fs.cstr.control_volume.energy_holdup[m.fs.time.first(), "aq"].fix(300)

    m.fs.mixer.E_inlet.conc_mol.fix(0)
    m.fs.mixer.S_inlet.conc_mol.fix(0)
    m.fs.mixer.E_inlet.conc_mol[:, "Solvent"].fix(1.0)
    m.fs.mixer.S_inlet.conc_mol[:, "Solvent"].fix(1.0)

    for t, j in m.fs.time * m.fs.properties.component_list:
        if j == "E":
            m.fs.mixer.E_inlet.conc_mol[t, j].fix(inlet_E)
        elif j == "S":
            m.fs.mixer.S_inlet.conc_mol[t, j].fix(inlet_S)

    m.fs.mixer.E_inlet.flow_vol.fix(0.1)
    m.fs.mixer.S_inlet.flow_vol.fix(2.1)

    # Specify a perturbation to substrate flow rate:
    for t in m.fs.time:
        if t < horizon / 4:
            continue
        else:
            m.fs.mixer.S_inlet.flow_vol[t].fix(3.0)

    m.fs.mixer.E_inlet.temperature.fix(290)
    m.fs.mixer.S_inlet.temperature.fix(310)

    m.fs.inlet = Arc(source=m.fs.mixer.outlet, destination=m.fs.cstr.inlet)

    # Fix "initial condition" for outlet flow rate, as here it cannot be
    # specified by the PID controller
    m.fs.cstr.outlet.flow_vol[m.fs.time.first()].fix(2.2)

    TransformationFactory("network.expand_arcs").apply_to(m.fs)

    return m


@pytest.mark.component
@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialize():
    """Very rough test, just to make sure degrees of freedom are not violated."""
    mod = make_model(horizon=2, ntfe=20, ntcp=1, inlet_E=11.91, inlet_S=12.92)

    assert degrees_of_freedom(mod) == 0

    originally_active = ComponentMap(
        [
            (comp, comp.active)
            for comp in mod.component_data_objects((Block, Constraint))
        ]
    )
    originally_fixed = ComponentMap(
        [(var, var.fixed) for var in mod.component_data_objects(Var)]
    )

    initialize_by_time_element(
        mod.fs, mod.fs.time, solver=solver, outlvl=idaeslog.DEBUG, fix_diff_only=False
    )
    assert degrees_of_freedom(mod) == 0

    for comp in mod.component_data_objects((Block, Constraint)):
        assert comp.active == originally_active[comp]
    for var in mod.component_data_objects(Var):
        assert var.fixed == originally_fixed[var]


if __name__ == "__main__":
    test_initialize()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# Import Pyomo libraries
import pyomo.environ as pyo
from idaes.core.util.model_statistics import degrees_of_freedom

import idaes.models_extra.power_generation.flowsheets.subcritical_power_plant.subcritical_power_plant as subcrit_plant
import idaes.models_extra.power_generation.flowsheets.subcritical_power_plant.steam_cycle_flowsheet as steam_cycle
import idaes.models_extra.power_generation.flowsheets.subcritical_power_plant.subcritical_boiler_flowsheet as blr
import idaes.models_extra.power_generation.flowsheets.subcritical_power_plant.subcritical_boiler as recyrc
from idaes.models.properties.general_helmholtz import helmholtz_available

import pytest


__author__ = "Boiler Subsystem Team (J. Ma, M. Zamarripa)"


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_subcritical_boiler_ss_build():
    m = blr.get_model(dynamic=False, init=False)
    assert degrees_of_freedom(m) == 12


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_subcritical_boiler_dynamic_build():
    m = blr.get_model(dynamic=True, init=False)
    assert degrees_of_freedom(m) == 223


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_subcritical_boiler():
    m = blr.main_steady_state()
    assert degrees_of_freedom(m) == 0
    # mass balance
    assert pytest.approx(0, abs=1e-2) == pyo.value(
        m.fs_main.fs_blr.aRH1.tube_inlet.flow_mol[0]
        + m.fs_main.fs_blr.aECON.tube_inlet.flow_mol[0]
        - m.fs_main.fs_blr.aPlaten.outlet.flow_mol[0]
        - m.fs_main.fs_blr.aRH2.tube_outlet.flow_mol[0]
        - m.fs_main.fs_blr.blowdown_split.FW_Blowdown.flow_mol[0]
        + m.fs_main.fs_blr.Attemp.Water_inlet.flow_mol[0]
    )
    # FEGT temperature
    assert pytest.approx(1399.96, rel=1e-5) == pyo.value(
        m.fs_main.fs_blr.aBoiler.flue_gas_outlet.temperature[0]
    )
    assert pytest.approx(329144395, rel=1e-5) == pyo.value(
        m.fs_main.fs_blr.aBoiler.heat_total_ww[0]
    )
    assert pytest.approx(418839867, rel=1e-5) == pyo.value(
        m.fs_main.fs_blr.aBoiler.heat_total[0]
    )


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_subcritical_boiler_dynamic():
    m = blr.main_dynamic()
    assert degrees_of_freedom(m) == 0
    assert pytest.approx(1399.9583, rel=1e-5) == pyo.value(
        m.fs_main.fs_blr.aBoiler.flue_gas_outlet.temperature[0]
    )
    assert pytest.approx(329144395, rel=1e-5) == pyo.value(
        m.fs_main.fs_blr.aBoiler.heat_total_ww[0]
    )
    assert pytest.approx(418839867, rel=1e-5) == pyo.value(
        m.fs_main.fs_blr.aBoiler.heat_total[0]
    )
    assert pytest.approx(1408.9636, rel=1e-5) == pyo.value(
        m.fs_main.fs_blr.aBoiler.flue_gas_outlet.temperature[60]
    )
    assert pytest.approx(334508402, rel=1e-5) == pyo.value(
        m.fs_main.fs_blr.aBoiler.heat_total_ww[60]
    )
    assert pytest.approx(426313190, rel=1e-5) == pyo.value(
        m.fs_main.fs_blr.aBoiler.heat_total[60]
    )


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_steam_cycle():
    m = steam_cycle.main_steady_state()
    assert degrees_of_freedom(m) == 0
    assert pytest.approx(61958, rel=1e-5) == pyo.value(
        m.fs_main.fs_stc.turb.inlet_split.inlet.enth_mol[0]
    )
    assert pytest.approx(12473.27146, rel=1e-5) == pyo.value(
        m.fs_main.fs_stc.turb.inlet_split.inlet.flow_mol[0]
    )
    assert pytest.approx(266.8806, rel=1e-5) == pyo.value(
        m.fs_main.fs_stc.power_output[0]
    )
    assert pytest.approx(0, abs=1e-2) == pyo.value(
        m.fs_main.fs_stc.turb.inlet_split.inlet.flow_mol[0]  # turbine inlet
        - m.fs_main.fs_stc.turb.hp_split[14].outlet_1.flow_mol[0]  # out to reheat
        + m.fs_main.fs_stc.turb.ip_stages[1].inlet.flow_mol[0]  # in from reheat
        - m.fs_main.fs_stc.spray_valve.outlet.flow_mol[0]  # out to attemperator
        - m.fs_main.fs_stc.fwh6.desuperheat.cold_side_outlet.flow_mol[
            0
        ]  # out to economizer
        + m.fs_main.fs_stc.makeup_valve.inlet.flow_mol[0]  # in from makeup
    )


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_subc_power_plant():
    m = subcrit_plant.main_steady_state()
    assert degrees_of_freedom(m) == 0
    assert pytest.approx(61634.3740, rel=1e-5) == pyo.value(
        m.fs_main.fs_stc.turb.inlet_split.inlet.enth_mol[0]
    )
    assert pytest.approx(14908.39189, rel=1e-5) == pyo.value(
        m.fs_main.fs_stc.turb.inlet_split.inlet.flow_mol[0]
    )
    assert pytest.approx(320, rel=1e-5) == pyo.value(m.fs_main.fs_stc.power_output[0])
    assert pytest.approx(0, abs=1e-2) == pyo.value(
        m.fs_main.fs_stc.turb.inlet_split.inlet.flow_mol[0]  # turbine inlet
        - m.fs_main.fs_stc.turb.hp_split[14].outlet_1.flow_mol[0]  # out to reheat
        + m.fs_main.fs_stc.turb.ip_stages[1].inlet.flow_mol[0]  # in from reheat
        - m.fs_main.fs_stc.spray_valve.outlet.flow_mol[0]  # out to attemperator
        - m.fs_main.fs_stc.fwh6.desuperheat.cold_side_outlet.flow_mol[
            0
        ]  # out to economizer
        + m.fs_main.fs_stc.makeup_valve.inlet.flow_mol[0]  # in from makeup
    )


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_dynamic_power_plant_build():
    # constructing and initializing dynamic power plant
    # not solving due to simulation time >20 min
    m = subcrit_plant.get_model(dynamic=True, init=False)
    assert m.dynamic is True
    # Lost 11 degrees of freedom because PIDControllers.mv_eqn[t0]
    # is no longer deactivated if calculate_initial_integral=False
    assert degrees_of_freedom(m) == 157


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_steadystate_power_plant_build():
    # constructing and initializing dynamic power plant
    # not solving due to simulation time >20 min
    m = subcrit_plant.get_model(dynamic=False, init=False)
    assert m.dynamic is False
    assert degrees_of_freedom(m) == -5


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_dynamic_steam_cycle():
    # constructing and initializing dynamic steam cycle flowsheet
    m = steam_cycle.get_model(dynamic=True)
    assert m.dynamic is True
    # Lost 7 degrees of freedom because PIDControllers.mv_eqn[t0]
    # is no longer deactivated if calculate_initial_integral=False
    assert degrees_of_freedom(m) == 0


@pytest.mark.usefixtures("run_in_tmp_path")
@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_subcritical_recirculation_system():
    m = recyrc.main()
    assert degrees_of_freedom(m) == 0



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Unit model to adjust size of streams to represent, for example, a stream being split across several identical units,
which are then all modeled as a single IDAES unit
"""
from functools import partial

from pyomo.environ import (
    Block,
    PositiveReals,
    units as pyunits,
    Var,
)
from pyomo.network import Port
from pyomo.common.config import ConfigBlock, ConfigValue, In

from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import (
    is_physical_parameter_block,
)
from idaes.core.base.var_like_expression import VarLikeExpression
from idaes.core.util.tables import create_stream_table_dataframe
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from idaes.models.unit_models.feed import FeedInitializer as StreamScalerInitializer

__author__ = "Douglas Allan, Tanner Polley"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("StreamScaler")
class StreamScalerData(UnitModelBlockData):
    """
    Unit model to adjust size of streams to represent, for example, a stream being split across several identical units,
    which are then all modeled as a single IDAES unit
    """

    default_initializer = StreamScalerInitializer

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
                **default** = False. Scaler blocks are always steady-state.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="Scaler blocks do not contain holdup, thus this must be False.",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for StreamScaler",
            doc="""Property parameter object used to define property
calculations, **default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property
block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """
        General build method for StreamScalerData. This method calls a number
        of sub-methods which automate the construction of expected attributes
        of unit models.

        Inheriting models should call `super().build`.

        Args:
            None

        Returns:
            None
        """
        # Call super.build()
        super(StreamScalerData, self).build()

        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = True

        # Call setup methods from ControlVolumeBlockData
        self._get_property_package()
        self._get_indexing_sets()

        self.properties = self.config.property_package.build_state_block(
            self.flowsheet().time, doc="Material properties at inlet", **tmp_dict
        )
        self.scaled_expressions = Block()
        self.multiplier = Var(
            initialize=1,
            domain=PositiveReals,
            units=pyunits.dimensionless,
            doc="Factor by which to scale dimensionless streams",
        )
        self.add_inlet_port(name="inlet", block=self.properties)
        self.outlet = Port(doc="Outlet port")

        def rule_scale_var(b, *args, var=None):
            return self.multiplier * var[args]

        def rule_no_scale_var(b, *args, var=None):
            return var[args]

        for var_name in self.inlet.vars.keys():
            var = getattr(self.inlet, var_name)
            if "flow" in var_name:
                rule = partial(rule_scale_var, var=var)
            else:
                rule = partial(rule_no_scale_var, var=var)
            self.scaled_expressions.add_component(
                var_name, VarLikeExpression(var.index_set(), rule=rule)
            )
            expr = getattr(self.scaled_expressions, var_name)
            self.outlet.add(expr, var_name)

    def initialize_build(
        blk, outlvl=idaeslog.NOTSET, optarg=None, solver=None, hold_state=False
    ):
        """
        Initialization routine for StreamScaler.

        Keyword Arguments:
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            hold_state : flag indicating whether the initialization routine
                     should unfix any state variables fixed during
                     initialization, **default** - False. **Valid values:**
                     **True** - states variables are not unfixed, and a dict of
                     returned containing flags for which states were fixed
                     during initialization, **False** - state variables are
                     unfixed after initialization by calling the release_state
                     method.

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization.
        """

        # Create solver

        # Initialize inlet state blocks
        flags = blk.properties.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        if hold_state is True:
            return flags
        else:
            blk.release_state(flags, outlvl=outlvl)

    def release_state(blk, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state = True.
            outlvl : sets output level of logging

        Returns:
            None
        """
        blk.properties.release_state(flags, outlvl=outlvl)

    def _get_stream_table_contents(self, time_point=0):
        io_dict = {
            "Inlet": self.inlet,
            # "Outlet": self.outlet,
        }
        return create_stream_table_dataframe(io_dict, time_point=time_point)

    def calculate_scaling_factors(self):
        # Scaling factors for the property block are calculated automatically
        super().calculate_scaling_factors()

        # Need to pass on scaling factors from the property block to the outlet
        # VarLikeExpressions so arcs get scaled right
        if self.multiplier.value == 0:
            default = 1
        else:
            default = 1 / self.multiplier.value

        scale = iscale.get_scaling_factor(
            self.multiplier, default=default, warning=False
        )
        for var_name in self.inlet.vars.keys():
            var = getattr(self.inlet, var_name)
            outlet_expr = getattr(self.outlet, var_name)
            for key, subvar in var.items():
                sf = iscale.get_scaling_factor(subvar, default=1, warning=True)
                iscale.set_scaling_factor(outlet_expr[key], scale * sf)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Interface for importing ONNX models into IDAES
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from enum import Enum
import numpy as np

from pyomo.common.dependencies import attempt_import

from idaes.core.surrogate.base.surrogate_base import SurrogateBase
from idaes.core.surrogate.sampling.scaling import OffsetScaler

keras, keras_available = attempt_import("tensorflow.keras")
omlt, omlt_available = attempt_import("omlt")

if omlt_available:
    from omlt import OmltBlock, OffsetScaling


class OMLTSurrogate(SurrogateBase):
    def __init__(
        self,
        input_labels,
        output_labels,
        input_bounds,
        input_scaler=None,
        output_scaler=None,
    ):
        """
        Standard SurrogateObject for surrogates based on Keras models.
        Utilizes the OMLT framework for importing Keras models to IDAES.

        Contains methods to both populate a Pyomo Block with constraints
        representing the surrogate and to evaluate the surrogate a set of user
        provided points.

        This constructor should only be used when first creating the surrogate within IDAES.
        Once created, this object can be stored to disk using save_to_folder and loaded
        with load_from_folder

        Args:
           onnx_model: Onnx model file to be loaded.
           input_labels: list of str
              The ordered list of labels corresponding to the inputs in the keras model
           output_labels: list of str
              The ordered list of labels corresponding to the outputs in the keras model
           input_bounds: None of dict of tuples
              Keys correspond to each of the input labels and values are the tuples of
              bounds (lb, ub)
           input_scaler: None or OffsetScaler
              The scaler to be used for the inputs. If None, then no scaler is used
           output_scaler: None of OffsetScaler
              The scaler to be used for the outputs. If None, then no scaler is used
        """
        super().__init__(
            input_labels=input_labels,
            output_labels=output_labels,
            input_bounds=input_bounds,
        )

        # make sure we are using the standard scaler
        if (
            input_scaler is not None
            and not isinstance(input_scaler, OffsetScaler)
            or output_scaler is not None
            and not isinstance(output_scaler, OffsetScaler)
        ):
            raise NotImplementedError("KerasSurrogate only supports the OffsetScaler.")

        # check that the input labels match
        if input_scaler is not None and input_scaler.expected_columns() != input_labels:
            raise ValueError(
                "KerasSurrogate created with input_labels that do not match"
                " the expected columns in the input_scaler.\n"
                "input_labels={}\n"
                "input_scaler.expected_columns()={}".format(
                    input_labels, input_scaler.expected_columns()
                )
            )

        # check that the output labels match
        if (
            output_scaler is not None
            and output_scaler.expected_columns() != output_labels
        ):
            raise ValueError(
                "KerasSurrogate created with output_labels that do not match"
                " the expected columns in the output_scaler.\n"
                "output_labels={}\n"
                "output_scaler.expected_columns()={}".format(
                    output_labels, output_scaler.expected_columns()
                )
            )

        self._input_scaler = input_scaler
        self._output_scaler = output_scaler

    class Formulation(Enum):
        FULL_SPACE = 1
        REDUCED_SPACE = 2
        RELU_BIGM = 3
        RELU_COMPLEMENTARITY = 4

    def generate_omlt_scaling_objecets(self):
        offset_inputs = np.zeros(self.n_inputs())
        factor_inputs = np.ones(self.n_inputs())
        offset_outputs = np.zeros(self.n_outputs())
        factor_outputs = np.ones(self.n_outputs())
        if self._input_scaler:
            offset_inputs = self._input_scaler.offset_series()[
                self.input_labels()
            ].to_numpy()
            factor_inputs = self._input_scaler.factor_series()[
                self.input_labels()
            ].to_numpy()
        if self._output_scaler:
            offset_outputs = self._output_scaler.offset_series()[
                self.output_labels()
            ].to_numpy()
            factor_outputs = self._output_scaler.factor_series()[
                self.output_labels()
            ].to_numpy()

        omlt_scaling = OffsetScaling(
            offset_inputs=offset_inputs,
            factor_inputs=factor_inputs,
            offset_outputs=offset_outputs,
            factor_outputs=factor_outputs,
        )

        # omlt takes *scaled* input bounds as a dictionary with int keys
        input_bounds = dict(enumerate(self.input_bounds().values()))
        scaled_input_bounds = omlt_scaling.get_scaled_input_expressions(input_bounds)
        scaled_input_bounds = {i: tuple(bnd) for i, bnd in scaled_input_bounds.items()}
        return omlt_scaling, scaled_input_bounds

    def populate_block_with_net(self, block, formulation_object):
        """
        Method to populate a Pyomo Block with the omlt model constraints and build its formulation.

        Args:
           block: Pyomo Block component
              The block to be populated with variables and/or constraints.
           formulation_object: omlt loaded network formulation
        """

        block.nn = OmltBlock()
        block.nn.build_formulation(
            formulation_object,
        )

        # input/output variables need to be constrained to be equal
        # auto-created variables that come from OMLT.
        input_idx_by_label = {s: i for i, s in enumerate(self._input_labels)}
        input_vars_as_dict = block.input_vars_as_dict()

        @block.Constraint(self._input_labels)
        def input_surrogate_ties(m, input_label):
            return (
                input_vars_as_dict[input_label]
                == block.nn.inputs[input_idx_by_label[input_label]]
            )

        output_idx_by_label = {s: i for i, s in enumerate(self._output_labels)}
        output_vars_as_dict = block.output_vars_as_dict()

        @block.Constraint(self._output_labels)
        def output_surrogate_ties(m, output_label):
            return (
                output_vars_as_dict[output_label]
                == block.nn.outputs[output_idx_by_label[output_label]]
            )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

__author__ = "Douglas Allan"

import pytest
import numpy as np

import pyomo.environ as pyo
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.models_extra.power_generation.unit_models.soc_submodels as soc
import idaes.models_extra.power_generation.unit_models.soc_submodels.testing as soc_testing


def common_components(nt, nz, ncomp):
    return {
        pyo.Var: {
            "temperature_z": nz * nt,
            "temperature_deviation_x0": nz * nt,
            "temperature_deviation_x": nz * nt,
            "temperature_deviation_x1": nz * nt,
            "enth_mol": nz * nt,
            "heat_transfer_coefficient": nt * nz,
            "velocity": nz * nt,
            "pressure": nz * nt,
            "flow_mol": nz * nt,
            "mole_frac_comp": nz * nt * ncomp,
            "conc_mol_comp": nz * nt * ncomp,
            "heat_flux_x0": nz * nt,
            "heat_flux_x1": nz * nt,
            "length_x": 1,
            "length_z": 1,
            "length_y": 1,
            "flow_mol_inlet": nt,
            "pressure_inlet": nt,
            "temperature_inlet": nt,
            "temperature_outlet": nt,
            "mole_frac_comp_inlet": nt * ncomp,
            "diff_eff_coeff": nt * nz * ncomp,
        },
        pyo.Constraint: {
            "flow_mol_eqn": nz * nt,
            "constant_pressure_eqn": nz * nt,
            "conc_mol_comp_eqn": nz * nt * ncomp,
            "enth_mol_eqn": nt * nz,
            "mole_frac_comp_eqn": nt * nz,
            "diff_eff_coeff_eqn": nt * nz * ncomp,
            "material_balance_eqn": nt * nz * ncomp,
            "energy_balance_eqn": nt * nz,
            "temperature_x0_eqn": nt * nz,
            "temperature_x1_eqn": nt * nz,
            "temperature_outlet_eqn": nt,
        },
        pyo.Expression: {
            "temperature_x0": nz * nt,
            "temperature": nz * nt,
            "temperature_x1": nz * nt,
            "flow_area": 1,
            "flow_mol_comp_inlet": nt * ncomp,
            "dz": nz,
            "node_volume": nz,
            "xface_area": nz,
            "enth_mol_inlet": nt,
            "vol_mol": nt * nz,
            "vol_mol_inlet": nt,
            "mass_transfer_coeff": nt * nz * ncomp,
            "material_flux_z_inlet": nt * ncomp,
            "enth_flux_z_inlet": nt,
            "material_flux_z": nt * (nz + 1) * ncomp,
            "material_flux_z_enth": nt * (nz + 1),
            "pressure_face": nt * (nz + 1),
            "material_balance_eqn": nt * nz * ncomp,
            "flow_mol_comp_outlet": nt * ncomp,
            "enth_mol_outlet": nt,
            # Pyomo considers these VarLikeExpressions as Expressions
            "flow_mol_outlet": nt,
            "pressure_outlet": nt,
            "mole_frac_comp_outlet": nt * ncomp,
        },
    }


def fix_boundary_conditions(channel):
    channel.heat_flux_x0.fix(0)
    channel.heat_flux_x1.fix(0)
    if channel.config.below_electrode:
        channel.material_flux_x1.fix(0)
    else:
        channel.material_flux_x0.fix(0)
    channel.flow_mol_inlet.fix()
    channel.temperature_inlet.fix()
    channel.pressure_inlet.fix()
    channel.mole_frac_comp_inlet.fix()


@pytest.fixture
def modelNoHoldup():
    time_set = [0, 1]
    zfaces = np.linspace(0, 1, 4).tolist()
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False, time_set=time_set, time_units=pyo.units.s)
    m.fs.oxygen_channel = soc.SocChannel(
        has_holdup=False,
        control_volume_zfaces=zfaces,
        opposite_flow=True,
        below_electrode=False,
        component_list=["O2", "H2O"],
    )
    m.fs.oxygen_channel.length_x.fix(0.002)
    m.fs.oxygen_channel.heat_transfer_coefficient.fix(100)
    m.fs.oxygen_channel.length_y.fix(0.08)
    m.fs.oxygen_channel.length_z.fix(0.08)
    m.fs.oxygen_channel.temperature_z.fix(1000)
    fix_boundary_conditions(m.fs.oxygen_channel)
    return m


@pytest.fixture
def modelHoldupNotDynamic():
    time_set = [0, 1]
    zfaces = np.linspace(0, 1, 6).tolist()
    m = soc_testing._cell_flowsheet_model(
        dynamic=False, time_set=time_set, zfaces=zfaces
    )

    m.fs.fuel_channel = soc.SocChannel(
        has_holdup=True,
        control_volume_zfaces=zfaces,
        length_z=m.fs.length_z,
        length_y=m.fs.length_y,
        temperature_z=m.fs.temperature_z,
        opposite_flow=False,
        below_electrode=True,
        component_list=["H2", "H2O", "N2"],
    )
    m.fs.fuel_channel.length_x.fix(0.002)
    m.fs.fuel_channel.heat_transfer_coefficient.fix(100)
    fix_boundary_conditions(m.fs.fuel_channel)
    return m


@pytest.mark.build
@pytest.mark.unit
def test_build_modelNoHoldup(modelNoHoldup):
    channel = modelNoHoldup.fs.oxygen_channel
    nz = len(channel.znodes)
    nt = len(channel.flowsheet().time)
    ncomp = len(channel.component_list)

    comp_dict = common_components(nt, nz, ncomp)
    comp_dict[pyo.Var]["conc_mol_comp_deviation_x0"] = nt * nz * ncomp
    comp_dict[pyo.Var]["material_flux_x0"] = nt * nz * ncomp
    comp_dict[pyo.Constraint]["material_flux_x0_eqn"] = nt * nz * ncomp
    comp_dict[pyo.Param] = {}
    comp_dict[pyo.Param]["dconc_mol_compdt"] = nt * nz * ncomp
    comp_dict[pyo.Param]["dcedt"] = nt * nz
    comp_dict[pyo.Param]["conc_mol_comp_deviation_x1"] = nt * nz * ncomp
    comp_dict[pyo.Param]["material_flux_x1"] = nt * nz * ncomp

    soc_testing._build_test_utility(
        block=channel,
        comp_dict=comp_dict,
    )
    assert degrees_of_freedom(channel) == 0


@pytest.mark.component
def test_units_no_holdup(modelNoHoldup):
    assert_units_consistent(modelNoHoldup)


@pytest.mark.build
@pytest.mark.unit
def test_build_modelHoldupNotDynamic(modelHoldupNotDynamic):
    channel = modelHoldupNotDynamic.fs.fuel_channel
    nz = len(channel.znodes)
    nt = len(channel.flowsheet().time)
    ncomp = len(channel.component_list)

    comp_dict = common_components(nt, nz, ncomp)
    comp_dict[pyo.Var]["conc_mol_comp_deviation_x1"] = nt * nz * ncomp
    comp_dict[pyo.Var]["material_flux_x1"] = nt * nz * ncomp
    comp_dict[pyo.Constraint]["material_flux_x1_eqn"] = nt * nz * ncomp
    comp_dict[pyo.Var]["int_energy_mol"] = nt * nz
    comp_dict[pyo.Var]["int_energy_density"] = nt * nz
    comp_dict[pyo.Param] = {}
    comp_dict[pyo.Param]["dconc_mol_compdt"] = nt * nz * ncomp
    comp_dict[pyo.Param]["dcedt"] = nt * nz
    comp_dict[pyo.Param]["conc_mol_comp_deviation_x0"] = nt * nz * ncomp
    comp_dict[pyo.Param]["material_flux_x0"] = nt * nz * ncomp
    comp_dict[pyo.Constraint]["int_energy_mol_eqn"] = nt * nz
    comp_dict[pyo.Constraint]["int_energy_density_eqn"] = nt * nz

    soc_testing._build_test_utility(
        block=channel,
        comp_dict=comp_dict,
        references=[
            "temperature_z",
            "length_z",
            "length_y",
        ],
    )
    assert degrees_of_freedom(channel) == 0


@pytest.mark.component
def test_units_holdup_not_dynamic(modelHoldupNotDynamic):
    assert_units_consistent(modelHoldupNotDynamic)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Benzene-Toluene phase equilibrium package using ideal liquid and vapor.

Example property package using the Generic Property Package Framework.
This exmample shows how to set up a property package to do benzene-toluene
phase equilibrium in the generic framework using ideal liquid and vapor
assumptions along with methods drawn from the pre-built IDAES property
libraries.
"""
# Import Pyomo units
from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import LiquidPhase, VaporPhase, Component
import idaes.logger as idaeslog

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    IdealBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.forms import fugacity
from idaes.models.properties.modular_properties.pure import Perrys
from idaes.models.properties.modular_properties.pure import RPP4


# Set up logger
_log = idaeslog.getLogger(__name__)


# ---------------------------------------------------------------------
# Configuration dictionary for an ideal Benzene-Toluene system

# Data Sources:
# [1] The Properties of Gases and Liquids (1987)
#     4th edition, Chemical Engineering Series - Robert C. Reid
# [2] Perry's Chemical Engineers' Handbook 7th Ed.
# [3] Engineering Toolbox, https://www.engineeringtoolbox.com
#     Retrieved 1st December, 2019

configuration = {
    # Specifying components
    "components": {
        "benzene": {
            "type": Component,
            "elemental_composition": {"C": 6, "H": 6},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP4,
            "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            "parameter_data": {
                "mw": (78.1136e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (48.9e5, pyunits.Pa),  # [1]
                "temperature_crit": (562.2, pyunits.K),  # [1]
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (1.0162, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.2655, None),
                    "3": (562.16, pyunits.K),
                    "4": (0.28212, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (-3.392e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (4.739e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-3.017e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (7.130e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (1.29e2, pyunits.J / pyunits.kmol / pyunits.K),  # [2]
                    "2": (-1.7e-1, pyunits.J / pyunits.kmol / pyunits.K**2),
                    "3": (6.48e-4, pyunits.J / pyunits.kmol / pyunits.K**3),
                    "4": (0, pyunits.J / pyunits.kmol / pyunits.K**4),
                    "5": (0, pyunits.J / pyunits.kmol / pyunits.K**5),
                },
                "enth_mol_form_liq_comp_ref": (49.0e3, pyunits.J / pyunits.mol),  # [3]
                "enth_mol_form_vap_comp_ref": (82.9e3, pyunits.J / pyunits.mol),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (-6.98273, None),  # [1]
                    "B": (1.33213, None),
                    "C": (-2.62863, None),
                    "D": (-3.33399, None),
                },
            },
        },
        "toluene": {
            "type": Component,
            "elemental_composition": {"C": 7, "H": 8},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP4,
            "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            "parameter_data": {
                "mw": (92.1405e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (41e5, pyunits.Pa),  # [1]
                "temperature_crit": (591.8, pyunits.K),  # [1]
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (0.8488, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.26655, None),
                    "3": (591.8, pyunits.K),
                    "4": (0.2878, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (-2.435e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (5.125e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-2.765e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (4.911e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (1.40e2, pyunits.J / pyunits.kmol / pyunits.K),  # [2]
                    "2": (-1.52e-1, pyunits.J / pyunits.kmol / pyunits.K**2),
                    "3": (6.95e-4, pyunits.J / pyunits.kmol / pyunits.K**3),
                    "4": (0, pyunits.J / pyunits.kmol / pyunits.K**4),
                    "5": (0, pyunits.J / pyunits.kmol / pyunits.K**5),
                },
                "enth_mol_form_liq_comp_ref": (12.0e3, pyunits.J / pyunits.mol),  # [3]
                "enth_mol_form_vap_comp_ref": (50.1e3, pyunits.J / pyunits.mol),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (-7.28607, None),  # [1]
                    "B": (1.38091, None),
                    "C": (-2.83433, None),
                    "D": (-2.79168, None),
                },
            },
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {"type": LiquidPhase, "equation_of_state": Ideal},
        "Vap": {"type": VaporPhase, "equation_of_state": Ideal},
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (273.15, 300, 450, pyunits.K),
        "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
    },
    "pressure_ref": (1e5, pyunits.Pa),
    "temperature_ref": (300, pyunits.K),
    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": IdealBubbleDew,
}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for autoscalers.

Author: Andrew Lee
"""
from math import sqrt
import pytest

from pyomo.environ import Block, ConcreteModel, Constraint, Objective, Set, Suffix, Var

from idaes.core.scaling import AutoScaler


@pytest.fixture
def model():
    m = ConcreteModel()
    m.s = Set(initialize=[1, 2, 3, 4])

    m.b = Block(m.s)

    m.v1 = Var(initialize=2)
    m.v2 = Var(m.s, initialize=10)

    for bd in m.b.values():
        bd.v3 = Var(initialize=10)
        bd.v4 = Var(initialize=100)

    @m.Constraint(m.s)
    def c1(blk, i):
        return blk.v1**i == blk.v2[i]

    @m.Constraint(m.s)
    def c2(blk, i):
        return blk.v2[i] == blk.b[i].v3

    for bd in m.b.values():

        @bd.Constraint()
        def c3(blk):
            return blk.v3**2 == blk.v4

    m.o = Objective(expr=m.v1)

    return m


class TestAutoscaleVarMagnitude:
    @pytest.mark.unit
    def test_var_data(self, model):
        scaler = AutoScaler()
        scaler.scale_variables_by_magnitude(model.v1)
        scaler.scale_variables_by_magnitude(model.v2[1])

        assert model.scaling_factor[model.v1] == pytest.approx(1 / 2, rel=1e-8)
        assert model.scaling_factor[model.v2[1]] == pytest.approx(1 / 10, rel=1e-8)
        assert len(model.scaling_factor) == 2

    @pytest.mark.unit
    def test_var_data_no_value(self):
        model = ConcreteModel()
        model.v1 = Var()

        scaler = AutoScaler()
        scaler.scale_variables_by_magnitude(model.v1)

        assert model.scaling_factor[model.v1] == 1

    @pytest.mark.unit
    def test_indexed_var(self, model):
        scaler = AutoScaler()
        scaler.scale_variables_by_magnitude(model.v2)

        for i in model.s:
            assert model.scaling_factor[model.v2[i]] == pytest.approx(1 / 10, rel=1e-8)
        assert len(model.scaling_factor) == 4

    @pytest.mark.unit
    def test_block_data(self, model):
        scaler = AutoScaler()
        scaler.scale_variables_by_magnitude(model.b[1])

        assert model.b[1].scaling_factor[model.b[1].v3] == pytest.approx(
            1 / 10, rel=1e-8
        )
        assert model.b[1].scaling_factor[model.b[1].v4] == pytest.approx(
            1 / 100, rel=1e-8
        )
        assert len(model.b[1].scaling_factor) == 2

    @pytest.mark.unit
    def test_indexed_block(self, model):
        scaler = AutoScaler()
        scaler.scale_variables_by_magnitude(model.b)

        for bd in model.b.values():
            sfx = bd.scaling_factor
            assert sfx[bd.v3] == pytest.approx(1 / 10, rel=1e-8)
            assert sfx[bd.v4] == pytest.approx(1 / 100, rel=1e-8)
            assert len(sfx) == 2

    @pytest.mark.unit
    def test_nested_blocks_descend(self, model):
        scaler = AutoScaler()
        scaler.scale_variables_by_magnitude(model)

        assert model.scaling_factor[model.v1] == pytest.approx(1 / 2, rel=1e-8)
        for i in model.s:
            assert model.scaling_factor[model.v2[i]] == pytest.approx(1 / 10, rel=1e-8)
        assert len(model.scaling_factor) == 5

        for bd in model.b.values():
            sfx = bd.scaling_factor
            assert sfx[bd.v3] == pytest.approx(1 / 10, rel=1e-8)
            assert sfx[bd.v4] == pytest.approx(1 / 100, rel=1e-8)
            assert len(sfx) == 2

    @pytest.mark.unit
    def test_nested_blocks_no_descend(self, model):
        scaler = AutoScaler()
        scaler.scale_variables_by_magnitude(model, descend_into=False)

        assert model.scaling_factor[model.v1] == pytest.approx(1 / 2, rel=1e-8)
        for i in model.s:
            assert model.scaling_factor[model.v2[i]] == pytest.approx(1 / 10, rel=1e-8)
        assert len(model.scaling_factor) == 5

        for bd in model.b.values():
            assert not hasattr(bd, "scaling_factor")

    @pytest.mark.unit
    def test_no_overwrite(self, model):
        # Add some scaling factors to ensure they are not overwritten
        model.scaling_factor = Suffix(direction=Suffix.EXPORT)
        model.scaling_factor[model.v1] = 20

        model.b[1].scaling_factor = Suffix(direction=Suffix.EXPORT)
        model.b[1].scaling_factor[model.b[1].v3] = 20

        scaler = AutoScaler(overwrite=False)
        scaler.scale_variables_by_magnitude(model)

        assert model.scaling_factor[model.v1] == 20
        for i in model.s:
            assert model.scaling_factor[model.v2[i]] == pytest.approx(1 / 10, rel=1e-8)
        assert len(model.scaling_factor) == 5

        for bd in model.b.values():
            sfx = bd.scaling_factor
            if bd is model.b[1]:
                assert sfx[bd.v3] == 20
            else:
                assert sfx[bd.v3] == pytest.approx(1 / 10, rel=1e-8)
            assert sfx[bd.v4] == pytest.approx(1 / 100, rel=1e-8)
            assert len(sfx) == 2

    @pytest.mark.unit
    def test_max_and_min(self, model):
        scaler = AutoScaler(
            max_variable_scaling_factor=1 / 10, min_variable_scaling_factor=1 / 50
        )
        scaler.scale_variables_by_magnitude(model)

        assert model.scaling_factor[model.v1] == pytest.approx(1 / 10, rel=1e-8)
        for i in model.s:
            assert model.scaling_factor[model.v2[i]] == pytest.approx(1 / 10, rel=1e-8)
        assert len(model.scaling_factor) == 5

        for bd in model.b.values():
            sfx = bd.scaling_factor
            assert sfx[bd.v3] == pytest.approx(1 / 10, rel=1e-8)
            assert sfx[bd.v4] == pytest.approx(1 / 50, rel=1e-8)
            assert len(sfx) == 2

    @pytest.mark.unit
    def test_var_fixed(self, model):
        model.v1.fix()
        scaler = AutoScaler()
        scaler.scale_variables_by_magnitude(model.v1)
        scaler.scale_variables_by_magnitude(model.v2[1])

        assert model.scaling_factor[model.v1] == pytest.approx(1 / 2, rel=1e-8)
        assert model.scaling_factor[model.v2[1]] == pytest.approx(1 / 10, rel=1e-8)
        assert len(model.scaling_factor) == 2

    @pytest.mark.unit
    def test_not_block_or_var(self, model):
        scaler = AutoScaler()

        with pytest.raises(TypeError, match="c1 is not a block or variable."):
            scaler.scale_variables_by_magnitude(model.c1)


class TestConstraintsByNorm:
    @pytest.mark.unit
    def test_not_block_or_var(self, model):
        scaler = AutoScaler()

        with pytest.raises(TypeError, match="v1 is not a block or constraint."):
            scaler.scale_constraints_by_jacobian_norm(model.v1)

    @pytest.mark.unit
    def test_block_data_L2(self, model):
        scaler = AutoScaler()

        scaler.scale_constraints_by_jacobian_norm(model.b[1])

        assert model.b[1].scaling_factor[model.b[1].c3] == pytest.approx(
            1 / sqrt(20**2 + 1**2), rel=1e-8
        )
        assert len(model.b[1].scaling_factor) == 1

    @pytest.mark.unit
    def test_block_data_L2_block_data(self, model):
        scaler = AutoScaler()

        scaler.scale_constraints_by_jacobian_norm(model.b)

        for bd in model.b.values():
            assert bd.scaling_factor[bd.c3] == pytest.approx(
                1 / sqrt(20**2 + 1**2), rel=1e-8
            )
            assert len(bd.scaling_factor) == 1

    @pytest.mark.unit
    def test_nested_blocks_L2(self, model):
        scaler = AutoScaler()

        scaler.scale_constraints_by_jacobian_norm(model)

        for i in model.s:
            assert model.scaling_factor[model.c1[i]] == pytest.approx(
                1 / sqrt((i * 2 ** (i - 1)) ** 2 + 1**2), rel=1e-8
            )
            assert model.scaling_factor[model.c2[i]] == pytest.approx(
                1 / sqrt(2), rel=1e-8
            )
        assert len(model.scaling_factor) == 8

        for bd in model.b.values():
            assert bd.scaling_factor[bd.c3] == pytest.approx(
                1 / sqrt(20**2 + 1**2), rel=1e-8
            )
            assert len(bd.scaling_factor) == 1

    @pytest.mark.unit
    def test_nested_blocks_L2_no_descent(self, model):
        scaler = AutoScaler()

        scaler.scale_constraints_by_jacobian_norm(model, descend_into=False)

        for i in model.s:
            assert model.scaling_factor[model.c1[i]] == pytest.approx(
                1 / sqrt((i * 2 ** (i - 1)) ** 2 + 1**2), rel=1e-8
            )
            assert model.scaling_factor[model.c2[i]] == pytest.approx(
                1 / sqrt(2), rel=1e-8
            )
        assert len(model.scaling_factor) == 8

        for bd in model.b.values():
            assert not hasattr(bd, "scaling_factor")

    @pytest.mark.unit
    def test_constraint_data_L2(self, model):
        scaler = AutoScaler()

        scaler.scale_constraints_by_jacobian_norm(model.c1[1])

        assert model.scaling_factor[model.c1[1]] == pytest.approx(
            1 / sqrt((1 * 2 ** (1 - 1)) ** 2 + 1**2), rel=1e-8
        )
        assert len(model.scaling_factor) == 1

    @pytest.mark.unit
    def test_indexed_constraint_L2(self, model):
        scaler = AutoScaler()

        scaler.scale_constraints_by_jacobian_norm(model.c1)

        for i in model.s:
            assert model.scaling_factor[model.c1[i]] == pytest.approx(
                1 / sqrt((i * 2 ** (i - 1)) ** 2 + 1**2), rel=1e-8
            )
        assert len(model.scaling_factor) == 4

    @pytest.mark.unit
    def test_block_data_L1(self, model):
        scaler = AutoScaler()

        scaler.scale_constraints_by_jacobian_norm(model.b[1], norm=1)

        assert model.b[1].scaling_factor[model.b[1].c3] == pytest.approx(
            1 / 21, rel=1e-8
        )
        assert len(model.b[1].scaling_factor) == 1

    @pytest.mark.unit
    def test_nested_blocks_L1(self, model):
        scaler = AutoScaler()

        scaler.scale_constraints_by_jacobian_norm(model, norm=1)

        for i in model.s:
            assert model.scaling_factor[model.c1[i]] == pytest.approx(
                1 / ((i * 2 ** (i - 1)) + 1), rel=1e-8
            )
            assert model.scaling_factor[model.c2[i]] == pytest.approx(1 / 2, rel=1e-8)
        assert len(model.scaling_factor) == 8

        for bd in model.b.values():
            assert bd.scaling_factor[bd.c3] == pytest.approx(1 / 21, rel=1e-8)
            assert len(bd.scaling_factor) == 1

    @pytest.mark.unit
    def test_constraint_data_L1(self, model):
        scaler = AutoScaler()

        scaler.scale_constraints_by_jacobian_norm(model.c1[1], norm=1)

        assert model.scaling_factor[model.c1[1]] == pytest.approx(
            1 / ((1 * 2 ** (1 - 1)) + 1), rel=1e-8
        )
        assert len(model.scaling_factor) == 1

    @pytest.mark.unit
    def test_indexed_constraint_L1(self, model):
        scaler = AutoScaler()

        scaler.scale_constraints_by_jacobian_norm(model.c1, norm=1)

        for i in model.s:
            assert model.scaling_factor[model.c1[i]] == pytest.approx(
                1 / ((i * 2 ** (i - 1)) + 1), rel=1e-8
            )
        assert len(model.scaling_factor) == 4

    @pytest.mark.unit
    def test_nested_blocks_L1_no_overwrite(self, model):
        # Add some scaling factors to ensure they are not overwritten
        model.scaling_factor = Suffix(direction=Suffix.EXPORT)
        model.scaling_factor[model.c1[1]] = 20

        model.b[1].scaling_factor = Suffix(direction=Suffix.EXPORT)
        model.b[1].scaling_factor[model.b[1].c3] = 20

        scaler = AutoScaler()

        scaler.scale_constraints_by_jacobian_norm(model, norm=1)

        for i in model.s:
            if i == 1:
                assert model.scaling_factor[model.c1[i]] == 20
            else:
                assert model.scaling_factor[model.c1[i]] == pytest.approx(
                    1 / ((i * 2 ** (i - 1)) + 1), rel=1e-8
                )
            assert model.scaling_factor[model.c2[i]] == pytest.approx(1 / 2, rel=1e-8)
        assert len(model.scaling_factor) == 8

        for k, bd in model.b.items():
            if k == 1:
                assert bd.scaling_factor[bd.c3] == 20
            else:
                assert bd.scaling_factor[bd.c3] == pytest.approx(1 / 21, rel=1e-8)
            assert len(bd.scaling_factor) == 1


class TestAutoScaleModel:
    @pytest.mark.unit
    def test_scale_model_default(self, model):
        scaler = AutoScaler()

        scaler.scale_model(model)

        c1_sf = {
            "c1[1]": 0.0980580676,
            "c1[2]": 0.0780868809,
            "c1[3]": 0.0384615385,
            "c1[4]": 0.0154376880,
        }

        for k, v in model.scaling_factor.items():
            if str(k).startswith("v2"):
                assert v == pytest.approx(0.1, rel=1e-8)
            elif str(k).startswith("v1"):
                assert v == pytest.approx(0.5, rel=1e-8)
            elif str(k).startswith("c2"):
                assert v == pytest.approx(0.5**0.5 * 0.1, rel=1e-8)
            elif str(k).startswith("c1"):
                assert v == pytest.approx(c1_sf[str(k)], rel=1e-8)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

__author__ = "Douglas Allan"

import pytest
import numpy as np

import pyomo.environ as pyo
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.models_extra.power_generation.unit_models.soc_submodels as soc
import idaes.models_extra.power_generation.unit_models.soc_submodels.testing as soc_testing


# Class for the electrolyte and interconnect
def common_components(nt, nz, nx):
    return {
        pyo.Var: {
            "temperature_z": nz * nt,
            "current_density": nz * nt,
            "length_z": 1,
            "length_y": 1,
            "temperature_deviation_x0": nz * nt,
            "heat_flux_x0": nz * nt,
            "temperature_deviation_x1": nz * nt,
            "heat_flux_x1": nz * nt,
            "temperature_deviation_x": nx * nz * nt,
            "length_x": 1,
            "resistivity_log_preexponential_factor": 1,
            "resistivity_thermal_exponent_dividend": 1,
            "heat_capacity": 1,
            "density": 1,
            "thermal_conductivity": 1,
        },
        pyo.Constraint: {
            "heat_flux_x0_eqn": nz * nt,
            "heat_flux_x1_eqn": nz * nt,
            "energy_balance_solid_eqn": nx * nz * nt,
        },
        pyo.Expression: {
            "temperature_x0": nz * nt,
            "temperature": nx * nz * nt,
            "temperature_x1": nz * nt,
            "dz": nz,
            "dx": nx,
            "node_volume": nx * nz,
            "zface_area": nx,
            "xface_area": nz,
            "current": nz * nt,
            "dTdx": (nx + 1) * nz * nt,
            "dTdz": (nz + 1) * nx * nt,
            "heat_flux_x": (nx + 1) * nz * nt,
            "heat_flux_z": (nz + 1) * nx * nt,
            "resistivity": nx * nz * nt,
            "resistance": nx * nz * nt,
            "voltage_drop": nx * nz * nt,
            "resistance_total": nz * nt,
            "voltage_drop_total": nz * nt,
            "joule_heating": nx * nz * nt,
        },
    }


@pytest.fixture
def modelNoHoldup():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False, time_set=[0, 1], time_units=pyo.units.s)
    m.fs.slab = soc.SocConductiveSlab(
        has_holdup=False,
        control_volume_zfaces=np.linspace(0, 1, 4).tolist(),
        control_volume_xfaces=np.linspace(0, 1, 5).tolist(),
    )
    slab = m.fs.slab

    m.fs.slab.length_x.fix(1.4e-4)
    m.fs.slab.heat_capacity.fix(470)
    m.fs.slab.density.fix(5160)
    m.fs.slab.thermal_conductivity.fix(2.16)
    m.fs.slab.resistivity_log_preexponential_factor.fix(pyo.log(1.07e-4))
    m.fs.slab.resistivity_thermal_exponent_dividend.fix(7237)

    slab.temperature_deviation_x0.fix(0)
    slab.heat_flux_x0.fix(0)
    slab.length_y.fix(0.08)
    slab.length_z.fix(0.08)
    slab.temperature_z.fix(1000)
    slab.current_density.fix(0)

    return m


@pytest.fixture
def modelHoldupNotDynamic():
    time_set = [0]
    zfaces = np.linspace(0, 1, 8).tolist()
    xfaces = np.linspace(0, 1, 12).tolist()
    m = soc_testing._cell_flowsheet_model(
        dynamic=False,
        time_set=time_set,
        zfaces=zfaces,
    )
    iznodes = m.fs.iznodes
    tset = m.fs.config.time
    m.fs.temperature_deviation_x0 = pyo.Var(
        tset, iznodes, initialize=0, units=pyo.units.K
    )
    m.fs.heat_flux_x0 = pyo.Var(
        tset, iznodes, initialize=0, units=pyo.units.W / pyo.units.m**2
    )
    m.fs.temperature_deviation_x1 = pyo.Var(
        tset, iznodes, initialize=0, units=pyo.units.K
    )
    m.fs.heat_flux_x1 = pyo.Var(
        tset, iznodes, initialize=0, units=pyo.units.W / pyo.units.m**2
    )

    m.fs.slab = soc.SocConductiveSlab(
        has_holdup=True,
        control_volume_zfaces=zfaces,
        control_volume_xfaces=xfaces,
        length_z=m.fs.length_z,
        length_y=m.fs.length_y,
        current_density=m.fs.current_density,
        temperature_z=m.fs.temperature_z,
        temperature_deviation_x0=m.fs.temperature_deviation_x0,
        heat_flux_x0=m.fs.heat_flux_x0,
        temperature_deviation_x1=m.fs.temperature_deviation_x1,
        heat_flux_x1=m.fs.heat_flux_x1,
    )
    m.fs.temperature_deviation_x0.fix(0)
    m.fs.heat_flux_x0.fix(0)

    m.fs.slab.length_x.fix(1.4e-4)
    m.fs.slab.heat_capacity.fix(470)
    m.fs.slab.density.fix(5160)
    m.fs.slab.thermal_conductivity.fix(2.16)
    m.fs.slab.resistivity_log_preexponential_factor.fix(pyo.log(1.07e-4))
    m.fs.slab.resistivity_thermal_exponent_dividend.fix(7237)

    return m


@pytest.mark.build
@pytest.mark.unit
def test_build_model_no_holdup(modelNoHoldup):
    slab = modelNoHoldup.fs.slab
    nx = len(slab.ixnodes)
    nz = len(slab.iznodes)
    nt = len(slab.flowsheet().time)
    comp_dict = common_components(nt, nz, nx)
    comp_dict[pyo.Param] = {}
    comp_dict[pyo.Param]["dcedt_solid"] = nt * nx * nz
    soc_testing._build_test_utility(
        slab,
        comp_dict=comp_dict,
    )

    assert degrees_of_freedom(slab) == 0


@pytest.mark.component
def test_units_no_holdup(modelNoHoldup):
    assert_units_consistent(modelNoHoldup)


@pytest.mark.build
@pytest.mark.unit
def test_build_model_holdup_not_dynamic(modelHoldupNotDynamic):
    slab = modelHoldupNotDynamic.fs.slab
    nx = len(slab.ixnodes)
    nz = len(slab.iznodes)
    nt = len(slab.flowsheet().time)
    comp_dict = common_components(nt, nz, nx)
    comp_dict[pyo.Var]["int_energy_density_solid"] = nx * nz * nt
    comp_dict[pyo.Param] = {}
    comp_dict[pyo.Param]["dcedt_solid"] = nt * nx * nz
    comp_dict[pyo.Constraint]["int_energy_density_solid_eqn"] = nt * nx * nz
    soc_testing._build_test_utility(
        slab,
        comp_dict=comp_dict,
        references=[
            "temperature_z",
            "current_density",
            "length_z",
            "length_y",
            "temperature_deviation_x0",
            "heat_flux_x0",
            "temperature_deviation_x1",
            "heat_flux_x1",
        ],
    )

    assert degrees_of_freedom(slab) == 0


@pytest.mark.component
def test_units_holdup_not_dynamic(modelHoldupNotDynamic):
    assert_units_consistent(modelHoldupNotDynamic)


@pytest.mark.component
def test_initialization_exception(modelNoHoldup):
    with pytest.raises(
        NotImplementedError,
        match="Initialization for the conductive_slab unit model is not implemented because there is no obvious "
        "set of boundary conditions to fix during solid_oxide_cell initialization, and it is not meant to be "
        "initialized in isolation.",
    ):
        modelNoHoldup.fs.slab.initialize_build()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest

from pyomo.environ import (
    ConcreteModel,
    Constraint,
    ExternalFunction,
    Expression,
    Param,
    Set,
    sqrt,
    Var,
)

from idaes.core import FlowsheetBlock, Component
import idaes.logger as idaeslog
from idaes.models.properties.cubic_eos.cubic_prop_pack import (
    CubicParameterBlock,
    CubicStateBlock,
    CubicEoS,
    EoS_param,
    CubicEoSInitializer,
)
from idaes.models.properties.modular_properties.eos.ceos import cubic_roots_available


# Set module level pyest marker
pytestmark = pytest.mark.cubic_root
prop_available = cubic_roots_available()


@pytest.mark.unit
def test_CubicEoS():
    assert len(CubicEoS) == 2
    assert CubicEoS.PR
    assert CubicEoS.SRK


class TestParameterBlock(object):
    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_build_default(self, caplog):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)
        with caplog.at_level(idaeslog.WARNING):
            m.fs.params = CubicParameterBlock()
        assert "May 2025 release." in caplog.text

        assert m.fs.params.state_block_class is CubicStateBlock
        assert m.fs.params.config.valid_phase == ("Vap", "Liq")

        assert isinstance(m.fs.params.phase_list, Set)
        assert len(m.fs.params.phase_list) == 2
        for p in m.fs.params.phase_list:
            assert p in ["Vap", "Liq"]

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_build_VL(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = CubicParameterBlock(valid_phase=("Vap", "Liq"))

        assert m.fs.params.state_block_class is CubicStateBlock
        assert m.fs.params.config.valid_phase == ("Vap", "Liq")

        assert isinstance(m.fs.params.phase_list, Set)
        assert len(m.fs.params.phase_list) == 2
        for p in m.fs.params.phase_list:
            assert p in ["Vap", "Liq"]

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_build_LV(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = CubicParameterBlock(valid_phase=("Liq", "Vap"))

        assert m.fs.params.state_block_class is CubicStateBlock
        assert m.fs.params.config.valid_phase == ("Liq", "Vap")

        assert isinstance(m.fs.params.phase_list, Set)
        assert len(m.fs.params.phase_list) == 2
        for p in m.fs.params.phase_list:
            assert p in ["Vap", "Liq"]

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_build_L(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = CubicParameterBlock(valid_phase="Liq")

        assert m.fs.params.state_block_class is CubicStateBlock
        assert m.fs.params.config.valid_phase == ("Liq")

        assert isinstance(m.fs.params.phase_list, Set)
        assert len(m.fs.params.phase_list) == 1
        for p in m.fs.params.phase_list:
            assert p in ["Liq"]

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_build_V(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = CubicParameterBlock(valid_phase="Vap")

        assert m.fs.params.state_block_class is CubicStateBlock
        assert m.fs.params.config.valid_phase == ("Vap")

        assert isinstance(m.fs.params.phase_list, Set)
        assert len(m.fs.params.phase_list) == 1
        for p in m.fs.params.phase_list:
            assert p in ["Vap"]


class TestStateBlock_LV_PR(object):
    @pytest.fixture()
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = CubicParameterBlock()

        m.fs.params.a = Component()
        m.fs.params.b = Component()

        m.fs.params.cubic_type = CubicEoS.PR

        m.fs.params.gas_const = Param(default=8.314462618)

        m.fs.params.pressure_crit = Param(
            m.fs.params.component_list, initialize={"a": 5e6, "b": 4e6}
        )
        m.fs.params.temperature_crit = Param(
            m.fs.params.component_list, initialize={"a": 500, "b": 600}
        )

        m.fs.params.omega = Param(
            m.fs.params.component_list, initialize={"a": 0.2, "b": 0.2}
        )

        m.fs.params.kappa = Param(
            m.fs.params.component_list,
            m.fs.params.component_list,
            initialize={
                ("a", "a"): 0.0,
                ("a", "b"): 0.0,
                ("b", "a"): 0.0,
                ("b", "b"): 0.0,
            },
        )

        return m

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_build_default(self, model):
        model.fs.props = model.fs.params.build_state_block([1])

        assert model.fs.props.default_initializer is CubicEoSInitializer

        assert isinstance(model.fs.props[1].flow_mol, Var)
        assert len(model.fs.props[1].flow_mol) == 1

        assert isinstance(model.fs.props[1].mole_frac_comp, Var)
        assert len(model.fs.props[1].mole_frac_comp) == 2
        for j in model.fs.props[1].mole_frac_comp:
            assert j in ["a", "b"]

        assert isinstance(model.fs.props[1].pressure, Var)
        assert len(model.fs.props[1].pressure) == 1
        assert isinstance(model.fs.props[1].temperature, Var)
        assert len(model.fs.props[1].temperature) == 1

        assert isinstance(model.fs.props[1].flow_mol_phase, Var)
        assert len(model.fs.props[1].flow_mol_phase) == 2
        for j in model.fs.props[1].flow_mol_phase:
            assert j in ["Vap", "Liq"]

        assert isinstance(model.fs.props[1].mole_frac_phase_comp, Var)
        assert len(model.fs.props[1].mole_frac_phase_comp) == 4
        for j in model.fs.props[1].mole_frac_phase_comp:
            assert j in [("Liq", "a"), ("Liq", "b"), ("Vap", "a"), ("Vap", "b")]

        assert isinstance(model.fs.props[1].total_flow_balance, Constraint)
        assert len(model.fs.props[1].total_flow_balance) == 1
        assert str(model.fs.props[1].total_flow_balance.body) == str(
            model.fs.props[1].flow_mol_phase["Liq"]
            + model.fs.props[1].flow_mol_phase["Vap"]
            - model.fs.props[1].flow_mol
        )

        assert isinstance(model.fs.props[1].component_flow_balances, Constraint)
        assert len(model.fs.props[1].component_flow_balances) == 2
        for j in model.fs.props[1].component_flow_balances:
            assert j in ["a", "b"]
            assert str(model.fs.props[1].component_flow_balances[j].body) == str(
                model.fs.props[1].flow_mol * model.fs.props[1].mole_frac_comp[j]
                - (
                    model.fs.props[1].flow_mol_phase["Liq"]
                    * model.fs.props[1].mole_frac_phase_comp["Liq", j]
                    + model.fs.props[1].flow_mol_phase["Vap"]
                    * model.fs.props[1].mole_frac_phase_comp["Vap", j]
                )
            )

        assert isinstance(model.fs.props[1].sum_mole_frac, Constraint)
        assert len(model.fs.props[1].sum_mole_frac) == 1
        assert str(model.fs.props[1].sum_mole_frac.body) == str(
            sum(
                model.fs.props[1].mole_frac_phase_comp["Liq", i]
                for i in model.fs.params.component_list
            )
            - sum(
                model.fs.props[1].mole_frac_phase_comp["Vap", i]
                for i in model.fs.params.component_list
            )
        )

        assert isinstance(model.fs.props[1].sum_mole_frac_out, Constraint)
        assert len(model.fs.props[1].sum_mole_frac_out) == 1
        assert str(model.fs.props[1].sum_mole_frac_out.body) == str(
            sum(
                model.fs.props[1].mole_frac_comp[i]
                for i in model.fs.params.component_list
            )
        )

        assert isinstance(model.fs.props[1]._teq, Var)
        assert len(model.fs.props[1]._teq) == 1

        assert isinstance(model.fs.props[1]._t1, Var)
        assert len(model.fs.props[1]._t1) == 1

        assert isinstance(model.fs.props[1].eps_1, Param)
        assert model.fs.props[1].eps_1.value == 0.01
        assert isinstance(model.fs.props[1].eps_2, Param)
        assert model.fs.props[1].eps_2.value == 0.0005

        assert isinstance(model.fs.props[1]._t1_constraint, Constraint)
        assert len(model.fs.props[1]._t1_constraint) == 1
        assert str(model.fs.props[1]._t1_constraint.body) == str(
            model.fs.props[1]._t1
            - 0.5
            * (
                model.fs.props[1].temperature
                + model.fs.props[1].temperature_bubble
                + sqrt(
                    (
                        model.fs.props[1].temperature
                        - model.fs.props[1].temperature_bubble
                    )
                    ** 2
                    + model.fs.props[1].eps_1 ** 2
                )
            )
        )

        assert isinstance(model.fs.props[1]._teq_constraint, Constraint)
        assert len(model.fs.props[1]._teq_constraint) == 1
        assert str(model.fs.props[1]._teq_constraint.body) == str(
            model.fs.props[1]._teq
            - 0.5
            * (
                model.fs.props[1]._t1
                + model.fs.props[1].temperature_dew
                - sqrt(
                    (model.fs.props[1]._t1 - model.fs.props[1].temperature_dew) ** 2
                    + model.fs.props[1].eps_2 ** 2
                )
            )
        )

        assert isinstance(model.fs.props[1]._tr_eq, Expression)
        assert len(model.fs.props[1]._tr_eq) == 2
        for j in model.fs.props[1]._tr_eq:
            assert j in ["a", "b"]
        assert str(model.fs.props[1]._tr_eq[j].expr) == str(
            model.fs.props[1]._teq / model.fs.props[1].params.temperature_crit[j]
        )

        assert isinstance(model.fs.props[1].equilibrium_constraint, Constraint)
        assert len(model.fs.props[1].equilibrium_constraint) == 2
        for j in model.fs.props[1].equilibrium_constraint:
            assert j in ["a", "b"]
            assert str(model.fs.props[1].equilibrium_constraint[j].body) == str(
                model.fs.props[1]._log_equilibrium_cubic("Vap", j)
                - model.fs.props[1]._log_equilibrium_cubic("Liq", j)
            )

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_common_cubic(self, model):
        model.fs.props = model.fs.params.build_state_block([1])

        eos = CubicEoS.PR

        assert model.fs.props[1].omegaA == EoS_param[eos]["omegaA"]
        assert model.fs.props[1].EoS_Bc == EoS_param[eos]["coeff_b"]
        assert model.fs.props[1].EoS_u == EoS_param[eos]["u"]
        assert model.fs.props[1].EoS_w == EoS_param[eos]["w"]
        assert model.fs.props[1].EoS_p == sqrt(
            EoS_param[eos]["u"] ** 2 - 4 * EoS_param[eos]["w"]
        )

        assert isinstance(model.fs.props[1].fw, Param)
        assert isinstance(model.fs.props[1].b, Param)

        assert isinstance(model.fs.props[1].compress_fact_liq_func, ExternalFunction)
        assert isinstance(model.fs.props[1].compress_fact_vap_func, ExternalFunction)


class TestStateBlock_L_PR(object):
    @pytest.fixture()
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = CubicParameterBlock(valid_phase="Liq")

        m.fs.params.a = Component()
        m.fs.params.b = Component()

        m.fs.params.cubic_type = CubicEoS.PR

        m.fs.params.gas_const = Param(default=8.314462618)

        m.fs.params.pressure_crit = Param(
            m.fs.params.component_list, initialize={"a": 5e6, "b": 4e6}
        )
        m.fs.params.temperature_crit = Param(
            m.fs.params.component_list, initialize={"a": 500, "b": 600}
        )

        m.fs.params.omega = Param(
            m.fs.params.component_list, initialize={"a": 0.2, "b": 0.2}
        )

        m.fs.params.kappa = Param(
            m.fs.params.component_list,
            m.fs.params.component_list,
            initialize={
                ("a", "a"): 0.0,
                ("a", "b"): 0.0,
                ("b", "a"): 0.0,
                ("b", "b"): 0.0,
            },
        )

        return m

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_build_default(self, model):
        model.fs.props = model.fs.params.build_state_block([1])

        assert isinstance(model.fs.props[1].flow_mol, Var)
        assert len(model.fs.props[1].flow_mol) == 1

        assert isinstance(model.fs.props[1].mole_frac_comp, Var)
        assert len(model.fs.props[1].mole_frac_comp) == 2
        for j in model.fs.props[1].mole_frac_comp:
            assert j in ["a", "b"]

        assert isinstance(model.fs.props[1].pressure, Var)
        assert len(model.fs.props[1].pressure) == 1
        assert isinstance(model.fs.props[1].temperature, Var)
        assert len(model.fs.props[1].temperature) == 1

        assert isinstance(model.fs.props[1].flow_mol_phase, Var)
        assert len(model.fs.props[1].flow_mol_phase) == 1
        for j in model.fs.props[1].flow_mol_phase:
            assert j in ["Liq"]

        assert isinstance(model.fs.props[1].mole_frac_phase_comp, Var)
        assert len(model.fs.props[1].mole_frac_phase_comp) == 2
        for j in model.fs.props[1].mole_frac_phase_comp:
            assert j in [("Liq", "a"), ("Liq", "b")]

        assert isinstance(model.fs.props[1].total_flow_balance, Constraint)
        assert len(model.fs.props[1].total_flow_balance) == 1
        assert str(model.fs.props[1].total_flow_balance.body) == str(
            model.fs.props[1].flow_mol - model.fs.props[1].flow_mol_phase["Liq"]
        )

        assert isinstance(model.fs.props[1].component_flow_balances, Constraint)
        assert len(model.fs.props[1].component_flow_balances) == 2
        for j in model.fs.props[1].component_flow_balances:
            assert j in ["a", "b"]
            assert str(model.fs.props[1].component_flow_balances[j].body) == str(
                model.fs.props[1].mole_frac_comp[j]
                - model.fs.props[1].mole_frac_phase_comp["Liq", j]
            )

        assert isinstance(model.fs.props[1].sum_mole_frac_out, Constraint)
        assert len(model.fs.props[1].sum_mole_frac_out) == 1
        assert str(model.fs.props[1].sum_mole_frac_out.body) == str(
            sum(
                model.fs.props[1].mole_frac_comp[i]
                for i in model.fs.params.component_list
            )
        )

        assert isinstance(model.fs.props[1]._teq, Expression)
        assert len(model.fs.props[1]._teq) == 1
        assert str(model.fs.props[1]._teq.expr) == str(model.fs.props[1].temperature)

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_common_cubic(self, model):
        model.fs.props = model.fs.params.build_state_block([1])

        eos = CubicEoS.PR

        assert model.fs.props[1].omegaA == EoS_param[eos]["omegaA"]
        assert model.fs.props[1].EoS_Bc == EoS_param[eos]["coeff_b"]
        assert model.fs.props[1].EoS_u == EoS_param[eos]["u"]
        assert model.fs.props[1].EoS_w == EoS_param[eos]["w"]
        assert model.fs.props[1].EoS_p == sqrt(
            EoS_param[eos]["u"] ** 2 - 4 * EoS_param[eos]["w"]
        )

        assert isinstance(model.fs.props[1].fw, Param)
        assert isinstance(model.fs.props[1].b, Param)

        assert isinstance(model.fs.props[1].compress_fact_liq_func, ExternalFunction)


class TestStateBlock_V_PR(object):
    @pytest.fixture()
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = CubicParameterBlock(valid_phase="Vap")

        m.fs.params.a = Component()
        m.fs.params.b = Component()

        m.fs.params.cubic_type = CubicEoS.PR

        m.fs.params.gas_const = Param(default=8.314462618)

        m.fs.params.pressure_crit = Param(
            m.fs.params.component_list, initialize={"a": 5e6, "b": 4e6}
        )
        m.fs.params.temperature_crit = Param(
            m.fs.params.component_list, initialize={"a": 500, "b": 600}
        )

        m.fs.params.omega = Param(
            m.fs.params.component_list, initialize={"a": 0.2, "b": 0.2}
        )

        m.fs.params.kappa = Param(
            m.fs.params.component_list,
            m.fs.params.component_list,
            initialize={
                ("a", "a"): 0.0,
                ("a", "b"): 0.0,
                ("b", "a"): 0.0,
                ("b", "b"): 0.0,
            },
        )

        return m

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_build_default(self, model):
        model.fs.props = model.fs.params.build_state_block([1])

        assert isinstance(model.fs.props[1].flow_mol, Var)
        assert len(model.fs.props[1].flow_mol) == 1

        assert isinstance(model.fs.props[1].mole_frac_comp, Var)
        assert len(model.fs.props[1].mole_frac_comp) == 2
        for j in model.fs.props[1].mole_frac_comp:
            assert j in ["a", "b"]

        assert isinstance(model.fs.props[1].pressure, Var)
        assert len(model.fs.props[1].pressure) == 1
        assert isinstance(model.fs.props[1].temperature, Var)
        assert len(model.fs.props[1].temperature) == 1

        assert isinstance(model.fs.props[1].flow_mol_phase, Var)
        assert len(model.fs.props[1].flow_mol_phase) == 1
        for j in model.fs.props[1].flow_mol_phase:
            assert j in ["Vap"]

        assert isinstance(model.fs.props[1].mole_frac_phase_comp, Var)
        assert len(model.fs.props[1].mole_frac_phase_comp) == 2
        for j in model.fs.props[1].mole_frac_phase_comp:
            assert j in [("Vap", "a"), ("Vap", "b")]

        assert isinstance(model.fs.props[1].total_flow_balance, Constraint)
        assert len(model.fs.props[1].total_flow_balance) == 1
        assert str(model.fs.props[1].total_flow_balance.body) == str(
            model.fs.props[1].flow_mol - model.fs.props[1].flow_mol_phase["Vap"]
        )

        assert isinstance(model.fs.props[1].component_flow_balances, Constraint)
        assert len(model.fs.props[1].component_flow_balances) == 2
        for j in model.fs.props[1].component_flow_balances:
            assert j in ["a", "b"]
            assert str(model.fs.props[1].component_flow_balances[j].body) == str(
                model.fs.props[1].mole_frac_comp[j]
                - model.fs.props[1].mole_frac_phase_comp["Vap", j]
            )

        assert isinstance(model.fs.props[1].sum_mole_frac_out, Constraint)
        assert len(model.fs.props[1].sum_mole_frac_out) == 1
        assert str(model.fs.props[1].sum_mole_frac_out.body) == str(
            sum(
                model.fs.props[1].mole_frac_comp[i]
                for i in model.fs.params.component_list
            )
        )

        assert isinstance(model.fs.props[1]._teq, Expression)
        assert len(model.fs.props[1]._teq) == 1
        assert str(model.fs.props[1]._teq.expr) == str(model.fs.props[1].temperature)

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_common_cubic(self, model):
        model.fs.props = model.fs.params.build_state_block([1])

        eos = CubicEoS.PR

        assert model.fs.props[1].omegaA == EoS_param[eos]["omegaA"]
        assert model.fs.props[1].EoS_Bc == EoS_param[eos]["coeff_b"]
        assert model.fs.props[1].EoS_u == EoS_param[eos]["u"]
        assert model.fs.props[1].EoS_w == EoS_param[eos]["w"]
        assert model.fs.props[1].EoS_p == sqrt(
            EoS_param[eos]["u"] ** 2 - 4 * EoS_param[eos]["w"]
        )

        assert isinstance(model.fs.props[1].fw, Param)
        assert isinstance(model.fs.props[1].b, Param)

        assert isinstance(model.fs.props[1].compress_fact_vap_func, ExternalFunction)


class TestStateBlock_LV_SRK(object):
    @pytest.fixture()
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = CubicParameterBlock()

        m.fs.params.a = Component()
        m.fs.params.b = Component()

        m.fs.params.cubic_type = CubicEoS.SRK

        m.fs.params.gas_const = Param(default=8.314462618)

        m.fs.params.pressure_crit = Param(
            m.fs.params.component_list, initialize={"a": 5e6, "b": 4e6}
        )
        m.fs.params.temperature_crit = Param(
            m.fs.params.component_list, initialize={"a": 500, "b": 600}
        )

        m.fs.params.omega = Param(
            m.fs.params.component_list, initialize={"a": 0.2, "b": 0.2}
        )

        m.fs.params.kappa = Param(
            m.fs.params.component_list,
            m.fs.params.component_list,
            initialize={
                ("a", "a"): 0.0,
                ("a", "b"): 0.0,
                ("b", "a"): 0.0,
                ("b", "b"): 0.0,
            },
        )

        return m

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_build_default(self, model):
        model.fs.props = model.fs.params.build_state_block([1])

        assert isinstance(model.fs.props[1].flow_mol, Var)
        assert len(model.fs.props[1].flow_mol) == 1

        assert isinstance(model.fs.props[1].mole_frac_comp, Var)
        assert len(model.fs.props[1].mole_frac_comp) == 2
        for j in model.fs.props[1].mole_frac_comp:
            assert j in ["a", "b"]

        assert isinstance(model.fs.props[1].pressure, Var)
        assert len(model.fs.props[1].pressure) == 1
        assert isinstance(model.fs.props[1].temperature, Var)
        assert len(model.fs.props[1].temperature) == 1

        assert isinstance(model.fs.props[1].flow_mol_phase, Var)
        assert len(model.fs.props[1].flow_mol_phase) == 2
        for j in model.fs.props[1].flow_mol_phase:
            assert j in ["Vap", "Liq"]

        assert isinstance(model.fs.props[1].mole_frac_phase_comp, Var)
        assert len(model.fs.props[1].mole_frac_phase_comp) == 4
        for j in model.fs.props[1].mole_frac_phase_comp:
            assert j in [("Liq", "a"), ("Liq", "b"), ("Vap", "a"), ("Vap", "b")]

        assert isinstance(model.fs.props[1].total_flow_balance, Constraint)
        assert len(model.fs.props[1].total_flow_balance) == 1
        assert str(model.fs.props[1].total_flow_balance.body) == str(
            model.fs.props[1].flow_mol_phase["Liq"]
            + model.fs.props[1].flow_mol_phase["Vap"]
            - model.fs.props[1].flow_mol
        )

        assert isinstance(model.fs.props[1].component_flow_balances, Constraint)
        assert len(model.fs.props[1].component_flow_balances) == 2
        for j in model.fs.props[1].component_flow_balances:
            assert j in ["a", "b"]
            assert str(model.fs.props[1].component_flow_balances[j].body) == str(
                model.fs.props[1].flow_mol * model.fs.props[1].mole_frac_comp[j]
                - (
                    model.fs.props[1].flow_mol_phase["Liq"]
                    * model.fs.props[1].mole_frac_phase_comp["Liq", j]
                    + model.fs.props[1].flow_mol_phase["Vap"]
                    * model.fs.props[1].mole_frac_phase_comp["Vap", j]
                )
            )

        assert isinstance(model.fs.props[1].sum_mole_frac, Constraint)
        assert len(model.fs.props[1].sum_mole_frac) == 1
        assert str(model.fs.props[1].sum_mole_frac.body) == str(
            sum(
                model.fs.props[1].mole_frac_phase_comp["Liq", i]
                for i in model.fs.params.component_list
            )
            - sum(
                model.fs.props[1].mole_frac_phase_comp["Vap", i]
                for i in model.fs.params.component_list
            )
        )

        assert isinstance(model.fs.props[1].sum_mole_frac_out, Constraint)
        assert len(model.fs.props[1].sum_mole_frac_out) == 1
        assert str(model.fs.props[1].sum_mole_frac_out.body) == str(
            sum(
                model.fs.props[1].mole_frac_comp[i]
                for i in model.fs.params.component_list
            )
        )

        assert isinstance(model.fs.props[1]._teq, Var)
        assert len(model.fs.props[1]._teq) == 1

        assert isinstance(model.fs.props[1]._t1, Var)
        assert len(model.fs.props[1]._t1) == 1

        assert isinstance(model.fs.props[1].eps_1, Param)
        assert model.fs.props[1].eps_1.value == 0.01
        assert isinstance(model.fs.props[1].eps_2, Param)
        assert model.fs.props[1].eps_2.value == 0.0005

        assert isinstance(model.fs.props[1]._t1_constraint, Constraint)
        assert len(model.fs.props[1]._t1_constraint) == 1
        assert str(model.fs.props[1]._t1_constraint.body) == str(
            model.fs.props[1]._t1
            - 0.5
            * (
                model.fs.props[1].temperature
                + model.fs.props[1].temperature_bubble
                + sqrt(
                    (
                        model.fs.props[1].temperature
                        - model.fs.props[1].temperature_bubble
                    )
                    ** 2
                    + model.fs.props[1].eps_1 ** 2
                )
            )
        )

        assert isinstance(model.fs.props[1]._teq_constraint, Constraint)
        assert len(model.fs.props[1]._teq_constraint) == 1
        assert str(model.fs.props[1]._teq_constraint.body) == str(
            model.fs.props[1]._teq
            - 0.5
            * (
                model.fs.props[1]._t1
                + model.fs.props[1].temperature_dew
                - sqrt(
                    (model.fs.props[1]._t1 - model.fs.props[1].temperature_dew) ** 2
                    + model.fs.props[1].eps_2 ** 2
                )
            )
        )

        assert isinstance(model.fs.props[1]._tr_eq, Expression)
        assert len(model.fs.props[1]._tr_eq) == 2
        for j in model.fs.props[1]._tr_eq:
            assert j in ["a", "b"]
        assert str(model.fs.props[1]._tr_eq[j].expr) == str(
            model.fs.props[1]._teq / model.fs.props[1].params.temperature_crit[j]
        )

        assert isinstance(model.fs.props[1].equilibrium_constraint, Constraint)
        assert len(model.fs.props[1].equilibrium_constraint) == 2
        for j in model.fs.props[1].equilibrium_constraint:
            assert j in ["a", "b"]
            assert str(model.fs.props[1].equilibrium_constraint[j].body) == str(
                model.fs.props[1]._log_equilibrium_cubic("Vap", j)
                - model.fs.props[1]._log_equilibrium_cubic("Liq", j)
            )

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_common_cubic(self, model):
        model.fs.props = model.fs.params.build_state_block([1])

        eos = CubicEoS.SRK

        assert model.fs.props[1].omegaA == EoS_param[eos]["omegaA"]
        assert model.fs.props[1].EoS_Bc == EoS_param[eos]["coeff_b"]
        assert model.fs.props[1].EoS_u == EoS_param[eos]["u"]
        assert model.fs.props[1].EoS_w == EoS_param[eos]["w"]
        assert model.fs.props[1].EoS_p == sqrt(
            EoS_param[eos]["u"] ** 2 - 4 * EoS_param[eos]["w"]
        )

        assert isinstance(model.fs.props[1].fw, Param)
        assert isinstance(model.fs.props[1].b, Param)

        assert isinstance(model.fs.props[1].compress_fact_liq_func, ExternalFunction)
        assert isinstance(model.fs.props[1].compress_fact_vap_func, ExternalFunction)


class TestStateBlock_L_SRK(object):
    @pytest.fixture()
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = CubicParameterBlock(valid_phase="Liq")

        m.fs.params.a = Component()
        m.fs.params.b = Component()

        m.fs.params.cubic_type = CubicEoS.SRK

        m.fs.params.gas_const = Param(default=8.314462618)

        m.fs.params.pressure_crit = Param(
            m.fs.params.component_list, initialize={"a": 5e6, "b": 4e6}
        )
        m.fs.params.temperature_crit = Param(
            m.fs.params.component_list, initialize={"a": 500, "b": 600}
        )

        m.fs.params.omega = Param(
            m.fs.params.component_list, initialize={"a": 0.2, "b": 0.2}
        )

        m.fs.params.kappa = Param(
            m.fs.params.component_list,
            m.fs.params.component_list,
            initialize={
                ("a", "a"): 0.0,
                ("a", "b"): 0.0,
                ("b", "a"): 0.0,
                ("b", "b"): 0.0,
            },
        )

        return m

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_build_default(self, model):
        model.fs.props = model.fs.params.build_state_block([1])

        assert isinstance(model.fs.props[1].flow_mol, Var)
        assert len(model.fs.props[1].flow_mol) == 1

        assert isinstance(model.fs.props[1].mole_frac_comp, Var)
        assert len(model.fs.props[1].mole_frac_comp) == 2
        for j in model.fs.props[1].mole_frac_comp:
            assert j in ["a", "b"]

        assert isinstance(model.fs.props[1].pressure, Var)
        assert len(model.fs.props[1].pressure) == 1
        assert isinstance(model.fs.props[1].temperature, Var)
        assert len(model.fs.props[1].temperature) == 1

        assert isinstance(model.fs.props[1].flow_mol_phase, Var)
        assert len(model.fs.props[1].flow_mol_phase) == 1
        for j in model.fs.props[1].flow_mol_phase:
            assert j in ["Liq"]

        assert isinstance(model.fs.props[1].mole_frac_phase_comp, Var)
        assert len(model.fs.props[1].mole_frac_phase_comp) == 2
        for j in model.fs.props[1].mole_frac_phase_comp:
            assert j in [("Liq", "a"), ("Liq", "b")]

        assert isinstance(model.fs.props[1].total_flow_balance, Constraint)
        assert len(model.fs.props[1].total_flow_balance) == 1
        assert str(model.fs.props[1].total_flow_balance.body) == str(
            model.fs.props[1].flow_mol - model.fs.props[1].flow_mol_phase["Liq"]
        )

        assert isinstance(model.fs.props[1].component_flow_balances, Constraint)
        assert len(model.fs.props[1].component_flow_balances) == 2
        for j in model.fs.props[1].component_flow_balances:
            assert j in ["a", "b"]
            assert str(model.fs.props[1].component_flow_balances[j].body) == str(
                model.fs.props[1].mole_frac_comp[j]
                - model.fs.props[1].mole_frac_phase_comp["Liq", j]
            )

        assert isinstance(model.fs.props[1].sum_mole_frac_out, Constraint)
        assert len(model.fs.props[1].sum_mole_frac_out) == 1
        assert str(model.fs.props[1].sum_mole_frac_out.body) == str(
            sum(
                model.fs.props[1].mole_frac_comp[i]
                for i in model.fs.params.component_list
            )
        )

        assert isinstance(model.fs.props[1]._teq, Expression)
        assert len(model.fs.props[1]._teq) == 1
        assert str(model.fs.props[1]._teq.expr) == str(model.fs.props[1].temperature)

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_common_cubic(self, model):
        model.fs.props = model.fs.params.build_state_block([1])

        eos = CubicEoS.SRK

        assert model.fs.props[1].omegaA == EoS_param[eos]["omegaA"]
        assert model.fs.props[1].EoS_Bc == EoS_param[eos]["coeff_b"]
        assert model.fs.props[1].EoS_u == EoS_param[eos]["u"]
        assert model.fs.props[1].EoS_w == EoS_param[eos]["w"]
        assert model.fs.props[1].EoS_p == sqrt(
            EoS_param[eos]["u"] ** 2 - 4 * EoS_param[eos]["w"]
        )

        assert isinstance(model.fs.props[1].fw, Param)
        assert isinstance(model.fs.props[1].b, Param)

        assert isinstance(model.fs.props[1].compress_fact_liq_func, ExternalFunction)


class TestStateBlock_V_SRK(object):
    @pytest.fixture()
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = CubicParameterBlock(valid_phase="Vap")

        m.fs.params.a = Component()
        m.fs.params.b = Component()

        m.fs.params.cubic_type = CubicEoS.SRK

        m.fs.params.gas_const = Param(default=8.314462618)

        m.fs.params.pressure_crit = Param(
            m.fs.params.component_list, initialize={"a": 5e6, "b": 4e6}
        )
        m.fs.params.temperature_crit = Param(
            m.fs.params.component_list, initialize={"a": 500, "b": 600}
        )

        m.fs.params.omega = Param(
            m.fs.params.component_list, initialize={"a": 0.2, "b": 0.2}
        )

        m.fs.params.kappa = Param(
            m.fs.params.component_list,
            m.fs.params.component_list,
            initialize={
                ("a", "a"): 0.0,
                ("a", "b"): 0.0,
                ("b", "a"): 0.0,
                ("b", "b"): 0.0,
            },
        )

        return m

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_build_default(self, model):
        model.fs.props = model.fs.params.build_state_block([1])

        assert isinstance(model.fs.props[1].flow_mol, Var)
        assert len(model.fs.props[1].flow_mol) == 1

        assert isinstance(model.fs.props[1].mole_frac_comp, Var)
        assert len(model.fs.props[1].mole_frac_comp) == 2
        for j in model.fs.props[1].mole_frac_comp:
            assert j in ["a", "b"]

        assert isinstance(model.fs.props[1].pressure, Var)
        assert len(model.fs.props[1].pressure) == 1
        assert isinstance(model.fs.props[1].temperature, Var)
        assert len(model.fs.props[1].temperature) == 1

        assert isinstance(model.fs.props[1].flow_mol_phase, Var)
        assert len(model.fs.props[1].flow_mol_phase) == 1
        for j in model.fs.props[1].flow_mol_phase:
            assert j in ["Vap"]

        assert isinstance(model.fs.props[1].mole_frac_phase_comp, Var)
        assert len(model.fs.props[1].mole_frac_phase_comp) == 2
        for j in model.fs.props[1].mole_frac_phase_comp:
            assert j in [("Vap", "a"), ("Vap", "b")]

        assert isinstance(model.fs.props[1].total_flow_balance, Constraint)
        assert len(model.fs.props[1].total_flow_balance) == 1
        assert str(model.fs.props[1].total_flow_balance.body) == str(
            model.fs.props[1].flow_mol - model.fs.props[1].flow_mol_phase["Vap"]
        )

        assert isinstance(model.fs.props[1].component_flow_balances, Constraint)
        assert len(model.fs.props[1].component_flow_balances) == 2
        for j in model.fs.props[1].component_flow_balances:
            assert j in ["a", "b"]
            assert str(model.fs.props[1].component_flow_balances[j].body) == str(
                model.fs.props[1].mole_frac_comp[j]
                - model.fs.props[1].mole_frac_phase_comp["Vap", j]
            )

        assert isinstance(model.fs.props[1].sum_mole_frac_out, Constraint)
        assert len(model.fs.props[1].sum_mole_frac_out) == 1
        assert str(model.fs.props[1].sum_mole_frac_out.body) == str(
            sum(
                model.fs.props[1].mole_frac_comp[i]
                for i in model.fs.params.component_list
            )
        )

        assert isinstance(model.fs.props[1]._teq, Expression)
        assert len(model.fs.props[1]._teq) == 1
        assert str(model.fs.props[1]._teq.expr) == str(model.fs.props[1].temperature)

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_common_cubic(self, model):
        model.fs.props = model.fs.params.build_state_block([1])

        eos = CubicEoS.SRK

        assert model.fs.props[1].omegaA == EoS_param[eos]["omegaA"]
        assert model.fs.props[1].EoS_Bc == EoS_param[eos]["coeff_b"]
        assert model.fs.props[1].EoS_u == EoS_param[eos]["u"]
        assert model.fs.props[1].EoS_w == EoS_param[eos]["w"]
        assert model.fs.props[1].EoS_p == sqrt(
            EoS_param[eos]["u"] ** 2 - 4 * EoS_param[eos]["w"]
        )

        assert isinstance(model.fs.props[1].fw, Param)
        assert isinstance(model.fs.props[1].b, Param)

        assert isinstance(model.fs.props[1].compress_fact_vap_func, ExternalFunction)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from idaes.models_extra.power_generation.unit_models.helm.splitter import HelmSplitter
from idaes.models_extra.power_generation.unit_models.helm.valve_steam import (
    HelmValve,
    ValveFunctionType,
)
from idaes.models_extra.power_generation.unit_models.helm.mixer import HelmMixer
from idaes.models.unit_models import MomentumMixingType
from idaes.models_extra.power_generation.unit_models.helm.compressor import (
    HelmIsentropicCompressor,
)
from idaes.models_extra.power_generation.unit_models.helm.turbine import (
    HelmIsentropicTurbine,
)
from idaes.models_extra.power_generation.unit_models.helm.pump import HelmPump
from idaes.models_extra.power_generation.unit_models.helm.turbine_inlet import (
    HelmTurbineInletStage,
)
from idaes.models_extra.power_generation.unit_models.helm.turbine_stage import (
    HelmTurbineStage,
)
from idaes.models_extra.power_generation.unit_models.helm.turbine_outlet import (
    HelmTurbineOutletStage,
)
from idaes.models_extra.power_generation.unit_models.helm.turbine_multistage import (
    HelmTurbineMultistage,
)
from idaes.models_extra.power_generation.unit_models.helm.condenser_ntu import (
    HelmNtuCondenser,
    HelmNtuCondenserData,
)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from abc import ABC, abstractmethod
from numbers import Real
import pandas as pd
import numpy as np
import idaes.logger as idaeslog

_logger = idaeslog.getLogger(__name__)


class ForecastError(Exception):
    """Error to indicate error with forecasters."""


class AbstractPriceForecaster(ABC):
    """
    The abstract class for price forecaster.
    """

    @abstractmethod
    def forecast_day_ahead_and_real_time_prices(
        self, date, hour, bus, horizon, n_samples
    ):
        """
        Forecast both day-ahead and real-time market prices.

        Arguments:
            date: intended date of the forecasts

            hour: intended hour of the forecasts

            bus: intended bus of the forecasts

            horizon: number of the time periods of the forecasts

            n_samples: number of the samples

        Returns:
            dict: day-ahead price forecasts

            dict: real-time price forecasts

        """

    @abstractmethod
    def forecast_real_time_prices(self, date, hour, bus, horizon, n_samples):
        """
        Forecast real-time market prices.

        Arguments:
            date: intended date of the forecasts

            hour: intended hour of the forecasts

            bus: intended bus of the forecasts

            horizon: number of the time periods of the forecasts

            n_samples: number of the samples

        Returns:
            dict: real-time price forecasts

        """

    @abstractmethod
    def forecast_day_ahead_prices(self, date, hour, bus, horizon, n_samples):
        """
        Forecast day-ahead market prices.

        Arguments:
            date: intended date of the forecasts

            hour: intended hour of the forecasts

            bus: intended bus of the forecasts

            horizon: number of the time periods of the forecasts

            n_samples: number of the samples

        Returns:
            dict: day-ahead price forecasts

        """


class AbstractPrescientPriceForecaster(AbstractPriceForecaster):
    """
    Abstract class for price forecasters that will interface with Prescient.
    """

    @abstractmethod
    def fetch_hourly_stats_from_prescient(self, prescient_hourly_stats):
        """
        This method fetches the hourly stats from Prescient to the price forecaster
        once the hourly stats are published.

        Arguments:
            prescient_hourly_stats: Prescient HourlyStats object.

        Returns:
            None
        """

    @abstractmethod
    def fetch_day_ahead_stats_from_prescient(self, uc_date, uc_hour, day_ahead_result):
        """
        This method fetches the day-ahead market to the price forecaster after the
        UC is solved from Prescient through the coordinator.

        Arguments:
            ruc_date: the date of the day-ahead market we bid into.

            ruc_hour: the hour the RUC is being solved in the day before.

            day_ahead_result: a Prescient RucPlan object.

        Returns:
            None
        """


class PlaceHolderForecaster(AbstractPrescientPriceForecaster):
    """
    This a placeholder for a real price forecaster. This placeholder can takes
    representative daily values and standard deviations for real-time and
    day-ahead prices to forecast prices in the future.
    """

    def __init__(
        self,
        daily_da_price_means: list,
        daily_rt_price_means: list,
        daily_da_price_stds: list,
        daily_rt_price_stds: list,
    ):
        """
        Initialize the PlaceHolderForecaster.

        Arguments:
            daily_da_price_means: list of day-ahead price means

            daily_rt_price_means: list of real-time price means

            daily_da_price_stds: list of price standard deviations

            daily_rt_price_stds: list of real-time price standard deviations

        Returns:
            None
        """

        self.daily_da_price_means = daily_da_price_means
        self.daily_rt_price_means = daily_rt_price_means
        self.daily_da_price_stds = daily_da_price_stds
        self.daily_rt_price_stds = daily_rt_price_stds

    def forecast_day_ahead_and_real_time_prices(
        self, date, hour, bus, horizon, n_samples
    ):
        """
        Forecast both day-ahead and real-time market prices.

        Arguments:
            date: intended date of the forecasts

            hour: intended hour of the forecasts

            bus: intended bus of the forecasts

            horizon: number of the time periods of the forecasts

            n_samples: number of the samples

        Returns:
            dict: day-ahead price forecasts

            dict: real-time price forecasts

        """

        rt_forecast = self.forecast_real_time_prices(
            date, hour, bus, horizon, n_samples
        )
        da_forecast = self.forecast_day_ahead_prices(
            date, hour, bus, horizon, n_samples
        )

        return da_forecast, rt_forecast

    def forecast_day_ahead_prices(self, date, hour, bus, horizon, n_samples):
        """
        Forecast day-ahead market prices.

        Arguments:
            date: intended date of the forecasts

            hour: intended hour of the forecasts

            bus: intended bus of the forecasts

            horizon: number of the time periods of the forecasts

            n_samples: number of the samples

        Returns:
            dict: day-ahead price forecasts

        """

        return self._forecast(
            means=self.daily_da_price_means,
            stds=self.daily_da_price_stds,
            hour=hour,
            horizon=horizon,
            n_samples=n_samples,
        )

    def forecast_real_time_prices(self, date, hour, bus, horizon, n_samples):
        """
        Forecast real-time market prices.

        Arguments:
            date: intended date of the forecasts

            hour: intended hour of the forecasts

            bus: intended bus of the forecasts

            horizon: number of the time periods of the forecasts

            n_samples: number of the samples

        Returns:
            dict: real-time price forecasts

        """

        return self._forecast(
            means=self.daily_rt_price_means,
            stds=self.daily_rt_price_stds,
            hour=hour,
            horizon=horizon,
            n_samples=n_samples,
        )

    def _forecast(self, means, stds, hour, horizon, n_samples):
        """
        Generate price forecasts.

        Arguments:
            means: list of price means

            stds: list of price standard deviations

            hour: intended hour of the forecasts

            horizon: number of the time periods of the forecasts

            n_samples: number of the samples

        Returns:
            dict: real-time price forecasts
        """

        corresponding_means = [means[t % 24] for t in range(hour, hour + horizon)]
        corresponding_stds = [stds[t % 24] for t in range(hour, hour + horizon)]

        forecasts_arr = np.random.normal(
            loc=corresponding_means, scale=corresponding_stds, size=(n_samples, horizon)
        )
        forecasts_arr[forecasts_arr < 0] = 0

        return {i: list(forecasts_arr[i]) for i in range(n_samples)}

    def fetch_hourly_stats_from_prescient(self, prescient_hourly_stats):
        """
        This method fetches the hourly stats from Prescient to the price forecaster
        once the hourly stats are published.

        Arguments:
            prescient_hourly_stats: Prescient HourlyStats object.

        Returns:
            None
        """

        return

    def fetch_day_ahead_stats_from_prescient(self, uc_date, uc_hour, day_ahead_result):
        """
        This method fetches the day-ahead market to the price forecaster after the
        UC is solved from Prescient through the coordinator.

        Arguments:
            ruc_date: the date of the day-ahead market we bid into.

            ruc_hour: the hour the RUC is being solved in the day before.

            day_ahead_result: a Prescient RucPlan object.

        Returns:
            None
        """

        return


class Backcaster(AbstractPrescientPriceForecaster):
    """
    Generate price forecasts by directly using historical prices.
    """

    def __init__(
        self, historical_da_prices, historical_rt_prices, max_historical_days=10
    ):
        """
        Initialize the Backcaster.

        Arguments:
            historical_da_prices: dictionary of list for historical hourly day-ahead prices

            historical_rt_prices: dictionary of list for historical hourly real-time prices

            max_historical_days: maximum number of days of price data to store on the instance

        Returns:
            None
        """

        self.max_historical_days = max_historical_days
        self.historical_da_prices = historical_da_prices
        self.historical_rt_prices = historical_rt_prices
        self._current_day_rt_prices = {bus: [] for bus in historical_da_prices}

    def _validate_input_historical_price(self, historical_price):
        """
        Validate input historical prices.

        Arguments:
            historical_price: dictionary of list for historical hourly prices

        Returns:
            None
        """

        if not isinstance(historical_price, dict):
            raise TypeError(
                "Given historical price is not an dictionary object. Dictionaries with bus name (str) as keys are expected."
            )

        if len(historical_price) == 0:
            raise ValueError("Given historical price is empty.")

        for b, price_list in historical_price.items():
            if not isinstance(price_list, list):
                raise TypeError(
                    f"Given historical price for bus {b} is not a list object. A list of historical prices is expected."
                )

            n_prices = len(price_list)

            if n_prices < 24:
                raise ValueError(
                    f"At least a day of the historical prices (24 entries) is needed. For bus {b}, only {n_prices} are provided."
                )

            if n_prices % 24 != 0:
                raise ValueError(
                    f"The number of prices for each bus should be a multiple of 24. But for bus {b}, {n_prices} are provided."
                )

            num_days = len(historical_price[b]) // 24
            if num_days > self.max_historical_days:
                _logger.warning(
                    f"The number of days in the input historical prices for bus {b} is greater than the max value {self.max_historical_days}."
                    f" Dropping the data for the first {num_days - self.max_historical_days} day(s)."
                )

                historical_price[b] = historical_price[b][
                    ((num_days - self.max_historical_days) * 24) :
                ]

    @property
    def max_historical_days(self):
        """
        Property getter for max_historical_days.

        Returns:
            int: max historical days
        """

        return self._max_historical_days

    @max_historical_days.setter
    def max_historical_days(self, value):
        """
        Property setter for max_historical_days (validate before setting).

        Args:
            value: intended value for max_historical_days

        Returns:
            None
        """

        if not isinstance(value, Real):
            raise TypeError(
                f"max_historical_days must be a number, but {type(value)} is provided."
            )

        value = int(value)

        if value < 1:
            raise ValueError(
                f"max_historical_days must be >= 1, but {value} is provided."
            )

        self._max_historical_days = value

    @property
    def historical_da_prices(self):
        """
        Property getter for historical_da_prices.

        Returns:
            dict: saved historical day-ahead prices
        """

        return self._historical_da_prices

    @historical_da_prices.setter
    def historical_da_prices(self, value):
        """
        Property setter for historical_da_prices (validate before setting).

        Args:
            value: intended value for historical_da_prices

        Returns:
            None
        """

        self._validate_input_historical_price(value)
        self._historical_da_prices = value

    @property
    def historical_rt_prices(self):
        """
        Property getter for historical_rt_prices.

        Returns:
            dict: saved historical real-time prices
        """

        return self._historical_rt_prices

    @historical_rt_prices.setter
    def historical_rt_prices(self, value):
        """
        Property setter for historical_rt_prices (validate before setting).

        Args:
            value: intended value for historical_rt_prices

        Returns:
            None
        """

        self._validate_input_historical_price(value)
        self._historical_rt_prices = value

    def forecast_day_ahead_and_real_time_prices(
        self, date, hour, bus, horizon, n_samples
    ):
        """
        Forecast both day-ahead and real-time market prices.

        Arguments:
            date: intended date of the forecasts

            hour: intended hour of the forecasts

            bus: intended bus of the forecasts

            horizon: number of the time periods of the forecasts

            n_samples: number of the samples

        Returns:
            dict: day-ahead price forecasts

            dict: real-time price forecasts

        """

        rt_forecast = self.forecast_real_time_prices(
            date, hour, bus, horizon, n_samples
        )
        da_forecast = self.forecast_day_ahead_prices(
            date, hour, bus, horizon, n_samples
        )

        return da_forecast, rt_forecast

    def forecast_real_time_prices(self, date, hour, bus, horizon, n_samples):
        """
        Forecast real-time market prices.

        Arguments:
            date: intended date of the forecasts

            hour: intended hour of the forecasts

            bus: intended bus of the forecasts

            horizon: number of the time periods of the forecasts

            n_samples: number of the samples

        Returns:
            dict: real-time price forecasts

        """

        return self._forecast(
            historical_price_dict=self.historical_rt_prices,
            market="real-time",
            date=date,
            hour=hour,
            bus=bus,
            horizon=horizon,
            n_samples=n_samples,
        )

    def forecast_day_ahead_prices(self, date, hour, bus, horizon, n_samples):
        """
        Forecast day-ahead market prices.

        Arguments:
            date: intended date of the forecasts

            hour: intended hour of the forecasts

            bus: intended bus of the forecasts

            horizon: number of the time periods of the forecasts

            n_samples: number of the samples

        Returns:
            dict: day-ahead price forecasts

        """

        return self._forecast(
            historical_price_dict=self.historical_da_prices,
            market="day-ahead",
            date=date,
            hour=0,
            bus=bus,
            horizon=horizon,
            n_samples=n_samples,
        )

    def _forecast(
        self, historical_price_dict, market, date, hour, bus, horizon, n_samples
    ):
        """
        Forecast energy market prices using historical prices.

        Arguments:

            historical_price_dict: the dictionary that holds the intended historical prices

            market: the market that the price forecast is for, e.g., day-ahead

            date: intended date of the forecasts

            hour: intended hour of the forecasts

            bus: intended bus of the forecasts

            horizon: number of the time periods of the forecasts

            n_samples: number of the samples

        Returns:
            dict: price forecasts

        """

        if bus not in historical_price_dict:
            raise ForecastError(f"No {bus} {market} price available.")

        historical_price_len = len(historical_price_dict[bus])
        n_days = historical_price_len // 24

        forecast = {}

        for i in range(n_samples):
            day_idx = n_days - (i % n_days) - 1

            forecast[i] = [
                historical_price_dict[bus][t % historical_price_len]
                for t in range(day_idx * 24 + hour, day_idx * 24 + hour + horizon)
            ]

        return forecast

    def fetch_hourly_stats_from_prescient(self, prescient_hourly_stats):
        """
        This method fetches the hourly real-time prices from Prescient and store
        them on the price forecaster, once they are published. When the stored historical
        data size has exceeded the specified upper bound, drop the oldest data.

        Arguments:
            prescient_hourly_stats: Prescient HourlyStats object.

        Returns:
            None
        """

        # save the newest rt prices
        for b, price_list in self._current_day_rt_prices.items():
            price_list.append(prescient_hourly_stats.observed_bus_LMPs[b])

        # update the historical
        for b in self._current_day_rt_prices:

            # if a full day's data is ready, get them ready for future forecasts
            if len(self._current_day_rt_prices[b]) >= 24:
                self._historical_rt_prices[b] += self._current_day_rt_prices[b]
                self._current_day_rt_prices[b] = []

            # drop oldest historical prices if total stored data exceeded the upper bound
            while len(self._historical_rt_prices[b]) // 24 > self.max_historical_days:
                self._historical_rt_prices[b] = self._historical_rt_prices[b][24:]

        return

    def fetch_day_ahead_stats_from_prescient(self, uc_date, uc_hour, day_ahead_result):
        """
        This method fetches the hourly day-ahead prices from Prescient and store
        them on the price forecaster, once they are published. When the stored historical
        data size has exceeded the specified upper bound, drop the oldest data.

        Arguments:
            ruc_date: the date of the day-ahead market we bid into.

            ruc_hour: the hour the RUC is being solved in the day before.

            day_ahead_result: a Prescient RucPlan object.

        Returns:
            None
        """

        for b in self._historical_da_prices:

            # save the newest da prices
            self._historical_da_prices[b] += [
                day_ahead_result.ruc_market.day_ahead_prices.get((b, t))
                for t in range(24)
            ]

            # drop oldest historical prices if total stored data exceeded the upper bound
            while len(self._historical_da_prices[b]) // 24 > self.max_historical_days:
                self._historical_da_prices[b] = self._historical_da_prices[b][24:]

        return


class PerfectForecaster(AbstractPrescientPriceForecaster):
    """
    PerfectForecaster is designed for the renewable-PEM Parameterized Bidder.
    To bid into the renewable-PEM, user need to forecaster the renewable generation at each time period.
    """

    def __init__(self, data_path_or_df):
        """
        Perfect forecaster that reads the data from a Dataframe containing:
         - {bus}-DALMP
         - {bus}-RTLMP
         - {bus}-DACF and {bus}-RTCF for renewable plants
         Perfect forecaster has no forecaster errors when forecasting the capacity factors
         unless the user add errors in the data_path_or_df

         Argument:
            data_path_or_df: str or pandas DataFrame,
                if it is a str, it is the path of the file that stores the data of capacity factors;
                if it is a DataFrame, that df should store the data of capacity factors.
        """
        if isinstance(data_path_or_df, str):
            self.data = pd.read_csv(
                data_path_or_df, index_col="Datetime", parse_dates=True
            )
        elif isinstance(data_path_or_df, pd.DataFrame):
            self.data = data_path_or_df
        else:
            raise ValueError(
                "The data_path_or_df should be pandas DataFrame or a string of the csv path"
            )

    def __getitem__(self, index):
        return self.data[index]

    def fetch_hourly_stats_from_prescient(self, prescient_hourly_stats):
        """
        Fetch the hourly stats from prescient.
        No need to have it here but forecaster should have this function to avoid an error from Coordinator.
        """
        pass

    def fetch_day_ahead_stats_from_prescient(self, uc_date, uc_hour, day_ahead_result):
        """
        Fetch the day-ahead stats from prescient.
        No need to have it here but forecaster should have this function to avoid an error from Coordinator.
        """
        pass

    def forecast_day_ahead_and_real_time_prices(
        self, date: str, hour: int, bus: str, horizon: int, _
    ):
        """
        Forecast day-ahead and real-time prices.
        Not necessary to forecast prices when using PEMParameterized bidder.
        But we build a function here for the future development.

        Arguments
            date: str,
                intended date of the forecasts.
            hour: int,
                intended hour of the forecasts.
            bus: str,
                intended bus of the forecasts.
            horizon: int,
                number of the time periods of the forecasts.

        Returns:
            da_forecast: numpyp array,
                the day-ahead price forecasts.
            rt_forecast: numpyp array,
                the real-time price forecasts.
        """
        rt_forecast = self.forecast_real_time_prices(date, hour, bus, horizon, _)
        da_forecast = self.forecast_day_ahead_prices(date, hour, bus, horizon, _)
        return da_forecast, rt_forecast

    def get_column_from_data(self, date: str, hour: int, horizon: int, col: str):
        """
        Get the data from the dataframe.

        Arguments
            date: str,
                intended date of the forecasts.
            hour: int,
                intended hour of the forecasts.
            horizon: int,
                number of the time periods of the forecasts.
            col: str,
                the name of the column.

        """
        datetime_index = pd.to_datetime(date) + pd.Timedelta(hours=hour)
        forecast = self.data[self.data.index >= datetime_index].head(horizon)
        values = forecast[col].values
        if len(values) < horizon:
            values = np.append(values, self.data[col].values[: horizon - len(values)])
        return values

    def forecast_day_ahead_prices(self, date, hour, bus, horizon, _):
        return self.get_column_from_data(date, hour, horizon, f"{bus}-DALMP")

    def forecast_real_time_prices(self, date, hour, bus, horizon, _):
        return self.get_column_from_data(date, hour, horizon, f"{bus}-RTLMP")

    def forecast_day_ahead_capacity_factor(self, date, hour, gen, horizon):
        return self.get_column_from_data(date, hour, horizon, f"{gen}-DACF")

    def forecast_real_time_capacity_factor(self, date, hour, gen, horizon):
        return self.get_column_from_data(date, hour, horizon, f"{gen}-RTCF")
